import unittest
import urllib
import os
import tarfile
import zipfile
import numpy as np
import PIL.Image
import tensorflow as tf
from tensorflow.core.framework import graph_pb2
import tfcoreml as tf_converter

TMP_MODEL_DIR = '/Users/kit/tmp/tfcoreml'
TEST_IMAGE = '/Users/kit/github/MMdnn/mmdnn/conversion/examples/data/seagull.jpg'

def _download_file(url):
  """Download the file.
  url - The URL address of the frozen file
  fname - Filename of the frozen TF graph in the url.
  """
  dir_path = TMP_MODEL_DIR
  if not os.path.exists(dir_path):
      os.makedirs(dir_path)

  k = url.rfind('/')
  fname = url[k+1:]
  fpath = os.path.join(dir_path, fname)

  ftype = None
  if url.endswith(".tar.gz") or url.endswith(".tgz"):
    ftype = 'tgz'
  elif url.endswith('.zip'):
    ftype = 'zip'

  if not os.path.exists(fpath):
    urllib.urlretrieve(url, fpath)
  if ftype == 'tgz':
    tar = tarfile.open(fpath)
    tar.extractall(dir_path)
    tar.close()
  elif ftype == 'zip':
    zip_ref = zipfile.ZipFile(fpath, 'r')
    zip_ref.extractall(dir_path)
    zip_ref.close()

def _compute_max_relative_error(x,y):
  rerror = 0
  index = 0
  for i in range(len(x)):
    den = max(1.0, np.abs(x[i]), np.abs(y[i]))
    if np.abs(x[i]/den - y[i]/den) > rerror:
      rerror = np.abs(x[i]/den - y[i]/den)
      index = i
  return rerror, index

def _compute_SNR(x,y):
  noise = x - y
  noise_var = np.sum(noise ** 2)/len(noise) + 1e-7
  signal_energy = np.sum(y ** 2)/len(y)
  max_signal_energy = np.amax(y ** 2)
  SNR = 10 * np.log10(signal_energy/noise_var)
  PSNR = 10 * np.log10(max_signal_energy/noise_var)
  return SNR, PSNR

def _load_image(path, resize_to=None):
  img = PIL.Image.open(path)
  if resize_to is not None:
    img = img.resize(resize_to, PIL.Image.ANTIALIAS)
  img_np = np.array(img).astype(np.float32)
  return img_np, img

def _generate_data(input_shape, mode = 'random',
                   scale = 2.0/255, bias = -1,
                   img_size = 256):
  """
  Generate some random data according to a shape.
  """
  if input_shape is None or len(input_shape) == 0:
    return 0.5
  if mode == 'zeros':
    X = np.zeros(input_shape)
  elif mode == 'ones':
    X = np.ones(input_shape)
  elif mode == 'linear':
    X = np.array(range(np.product(input_shape))).reshape(input_shape)*1.0
  elif mode == 'random':
    X = np.random.rand(*input_shape)
  elif mode == 'random_zero_mean':
    X = np.random.rand(*input_shape)-0.5
  elif mode == 'image':
    # Load a real image and do default tf imageNet preprocessing
    img_np, _ = _load_image(TEST_IMAGE ,resize_to=(img_size, img_size))
    img_tf = np.expand_dims(img_np, axis = 0)
    X = img_tf * scale + bias
  elif mode == 'onehot_0':
    X = np.zeros(input_shape)
    X[0] = 1
  return X

def _tf_transpose(x, is_sequence=False):
  if not hasattr(x, "shape"):
    return x
  if len(x.shape) == 4:
    # [Batch, Height, Width, Channels] --> [Batch, Channels, Height, Width]
    x = np.transpose(x, [0,3,1,2])
    return np.expand_dims(x, axis=0)
  elif len(x.shape) == 3:
    # We only deal with non-recurrent networks for now
    # (H,W,C) --> (C,H,W)
    return np.transpose(x, [2,0,1])
  elif len(x.shape) == 2:
    if is_sequence:  # (N,S) --> (S,N,1,)
      return x.reshape(x.shape[::-1] + (1,))
    else:  # (N,C) --> (N,C,1,1)
      return x.reshape((1, ) + x.shape) # Dense
  elif len(x.shape) == 1:
    if is_sequence: # (S) --> (S,N,1,1,1)
      return x.reshape((x.shape[0], 1, 1, 1, 1))
    else:
      return x
  else:
    return x

class CorrectnessTest(unittest.TestCase):

  @classmethod
  def setUpClass(self):
    """ Set up the unit test by loading common utilities.
    """
    self.err_thresh = 0.15
    self.snr_thresh = 12
    self.psnr_thresh = 30
    self.red_bias = -1
    self.blue_bias = -1
    self.green_bias = -1
    self.image_scale = 2.0/255

  def _compare_tf_coreml_outputs(self, tf_out, coreml_out):
    self.assertEquals(len(tf_out), len(coreml_out))
    error, ind = _compute_max_relative_error(coreml_out, tf_out)
    SNR, PSNR = _compute_SNR(coreml_out, tf_out)
    self.assertGreater(SNR, self.snr_thresh)
    self.assertGreater(PSNR, self.psnr_thresh)
    self.assertLess(error, self.err_thresh)


  def _test_tf_model(self, tf_model_path, coreml_model, input_tensors,
      output_tensor_names, data_modes = 'random', delta = 1e-2,
      use_cpu_only = False, scale = 2.0/255, bias = -1,
      img_size = None, sequence_inputs = None):
    """ Common entry to testing routine (Tensors in, tensors out).
    tf_model_path - frozen TF model path
    coreml_model - MLModel object
    input_tensors -  list of (name,shape) for each input (placeholder)
    output_tensor_names - output_tensor_names, a list of strings
    sequence_inputs - dict of input names that are sequences for CoreML input
    """
    # Load TensorFlow model
    tf.reset_default_graph()
    graph_def = graph_pb2.GraphDef()
    with open(tf_model_path, "rb") as f:
        graph_def.ParseFromString(f.read())
    g = tf.import_graph_def(graph_def)

    if type(data_modes) is str:
      data_modes = [data_modes] * len(input_tensors)

    with tf.Session(graph = g) as sess:
      # Prepare inputs
      feed_dict = {}
      for idx, in_tensor in enumerate(input_tensors):
        ts_name, ts_shape = in_tensor
        ts_name = 'import/' + ts_name
        feed_dict[ts_name] = _generate_data(ts_shape,
                                  mode = data_modes[idx],
                                  scale = scale, bias = bias,
                                  img_size = img_size)
      # Run TF session
      out_tf_names = []
      for out_name in output_tensor_names:
        out_tf_names.append('import/' + out_name)
      result = sess.run(out_tf_names, feed_dict=feed_dict)

    # Evaluate coreml model
    coreml_inputs = {}
    for idx, in_tensor in enumerate(input_tensors):
      in_tensor_name, in_shape = in_tensor
      coreml_in_name = in_tensor_name.replace(':', '__').replace('/', '__')
      if in_tensor_name in sequence_inputs:
        coreml_inputs[coreml_in_name] = _tf_transpose(
            feed_dict['import/'+in_tensor_name], is_sequence=True).copy()
      else:
        coreml_inputs[coreml_in_name] = _tf_transpose(
          feed_dict['import/'+in_tensor_name]).copy()

    coreml_output = coreml_model.predict(coreml_inputs, useCPUOnly=use_cpu_only)

    for idx, out_name in enumerate(output_tensor_names):
      out_tensor_name = out_name.replace(':', '__').replace('/', '__')
      tp = _tf_transpose(result[idx]).flatten()
      cp = coreml_output[out_tensor_name].flatten()
      error, index = _compute_max_relative_error(tp, cp)
      snr, psnr = _compute_SNR(tp, cp)
      self._compare_tf_coreml_outputs(tp, cp)


  def _test_coreml_model_image_input(self, tf_model_path, coreml_model,
      input_tensor_name, output_tensor_name, img_size, useCPUOnly = False):
    """Test single image input conversions.
    tf_model_path - the TF model
    coreml_model - converted CoreML model
    input_tensor_name - the input image tensor name
    output_tensor_name - the output tensor name
    img_size - size of the image
    """

    img_np, img = _load_image(TEST_IMAGE ,resize_to=(img_size, img_size))
    img_tf = np.expand_dims(img_np, axis = 0)
    img_tf[:,:,:,0] = self.image_scale * img_tf[:,:,:,0] + self.red_bias
    img_tf[:,:,:,1] = self.image_scale * img_tf[:,:,:,1] + self.green_bias
    img_tf[:,:,:,2] = self.image_scale * img_tf[:,:,:,2] + self.blue_bias

    #evaluate the TF model
    tf.reset_default_graph()
    graph_def = graph_pb2.GraphDef()
    with open(tf_model_path, "rb") as f:
        graph_def.ParseFromString(f.read())
    g = tf.import_graph_def(graph_def)
    with tf.Session(graph=g) as sess:
      image_input_tensor = sess.graph.get_tensor_by_name('import/' + input_tensor_name)
      output = sess.graph.get_tensor_by_name('import/' + output_tensor_name)
      tf_out = sess.run(output,feed_dict={image_input_tensor: img_tf})
    if len(tf_out.shape) == 4:
      tf_out = np.transpose(tf_out, (0,3,1,2))
    tf_out_flatten = tf_out.flatten()

    #evaluate CoreML
    coreml_input_name = input_tensor_name.replace(':', '__').replace('/', '__')
    coreml_output_name = output_tensor_name.replace(':', '__').replace('/', '__')
    coreml_input = {coreml_input_name: img}

    #Test the default CoreML evaluation
    coreml_out = coreml_model.predict(coreml_input, useCPUOnly = useCPUOnly)[coreml_output_name]
    coreml_out_flatten = coreml_out.flatten()
    self._compare_tf_coreml_outputs(tf_out_flatten, coreml_out_flatten)

class TestModels(CorrectnessTest):

  def test_inception_v3_slim(self):
    #Download model
    url = 'https://storage.googleapis.com/download.tensorflow.org/models/inception_v3_2016_08_28_frozen.pb.tar.gz'
    tf_model_dir = _download_file(url = url)
    tf_model_path = os.path.join(TMP_MODEL_DIR, 'inception_v3_2016_08_28_frozen.pb')

    #Convert to coreml
    mlmodel_path = os.path.join(TMP_MODEL_DIR, 'inception_v3_2016_08_28.mlmodel')
    mlmodel = tf_converter.convert(
        tf_model_path = tf_model_path,
        mlmodel_path = mlmodel_path,
        output_feature_names = ['InceptionV3/Predictions/Softmax:0'],
        input_name_shape_dict = {'input:0':[1,299,299,3]},
        image_input_names = ['input:0'],
        red_bias = -1,
        green_bias = -1,
        blue_bias = -1,
        image_scale = 2.0/255.0)

    #Test predictions on an image
    self._test_coreml_model_image_input(
        tf_model_path = tf_model_path,
        coreml_model = mlmodel,
        input_tensor_name = 'input:0',
        output_tensor_name = 'InceptionV3/Predictions/Softmax:0',
        img_size = 299)

  def test_googlenet_v1_nonslim(self):
    #Download model
    url = 'https://storage.googleapis.com/download.tensorflow.org/models/inception5h.zip'
    tf_model_dir = _download_file(url = url)
    tf_model_path = os.path.join(TMP_MODEL_DIR, 'tensorflow_inception_graph.pb')

    #Convert to coreml
    mlmodel_path = os.path.join(TMP_MODEL_DIR, 'googlenet_v1_nonslim.mlmodel')
    mlmodel = tf_converter.convert(
        tf_model_path = tf_model_path,
        mlmodel_path = mlmodel_path,
        output_feature_names = ['softmax2:0'],
        input_name_shape_dict = {'input:0':[1,224,224,3]},
        image_input_names = ['input:0'],
        red_bias = -1,
        green_bias = -1,
        blue_bias = -1,
        image_scale = 2.0/255.0)

    #Test predictions on an image
    self._test_coreml_model_image_input(
        tf_model_path = tf_model_path,
        coreml_model = mlmodel,
        input_tensor_name = 'input:0',
        output_tensor_name = 'softmax2:0',
        img_size = 224)

  def test_googlenet_resnet_v2(self):
    url = 'https://storage.googleapis.com/download.tensorflow.org/models/inception_resnet_v2_2016_08_30_frozen.pb.tar.gz'
    tf_model_dir = _download_file(url = url)
    tf_model_path = os.path.join(TMP_MODEL_DIR, 'inception_resnet_v2_2016_08_30_frozen.pb')

    mlmodel_path = os.path.join(TMP_MODEL_DIR, 'inception_resnet_v2_2016_08_30_frozen.mlmodel')
    mlmodel = tf_converter.convert(
        tf_model_path = tf_model_path,
        mlmodel_path = mlmodel_path,
        output_feature_names = ['InceptionResnetV2/Logits/Predictions:0'],
        input_name_shape_dict = {'input:0':[1,299,299,3]},
        image_input_names = ['input:0'],
        red_bias = -1,
        green_bias = -1,
        blue_bias = -1,
        image_scale = 2.0/255.0)

    #Test predictions on an image
    self._test_coreml_model_image_input(
        tf_model_path = tf_model_path,
        coreml_model = mlmodel,
        input_tensor_name = 'input:0',
        output_tensor_name = 'InceptionResnetV2/Logits/Predictions:0',
        img_size = 299)

  def test_googlenet_v1_slim(self):
    url = 'https://storage.googleapis.com/download.tensorflow.org/models/inception_v1_2016_08_28_frozen.pb.tar.gz'
    tf_model_dir = _download_file(url = url)
    tf_model_path = os.path.join(TMP_MODEL_DIR, 'inception_v1_2016_08_28_frozen.pb')

    mlmodel_path = os.path.join(TMP_MODEL_DIR, 'inception_v1_2016_08_28_frozen.mlmodel')
    mlmodel = tf_converter.convert(
        tf_model_path = tf_model_path,
        mlmodel_path = mlmodel_path,
        output_feature_names = ['InceptionV1/Logits/Predictions/Softmax:0'],
        input_name_shape_dict = {'input:0':[1,244,224,3]},
        image_input_names = ['input:0'],
        red_bias = -1,
        green_bias = -1,
        blue_bias = -1,
        image_scale = 2.0/255.0)

    #Test predictions on an image
    self._test_coreml_model_image_input(
        tf_model_path = tf_model_path,
        coreml_model = mlmodel,
        input_tensor_name = 'input:0',
        output_tensor_name = 'InceptionV1/Logits/Predictions/Softmax:0',
        img_size = 224)

  def test_googlenet_v2_slim(self):
    url = 'https://storage.googleapis.com/download.tensorflow.org/models/inception_v2_2016_08_28_frozen.pb.tar.gz'
    tf_model_dir = _download_file(url = url)
    tf_model_path = os.path.join(TMP_MODEL_DIR, 'inception_v2_2016_08_28_frozen.pb')

    mlmodel_path = os.path.join(TMP_MODEL_DIR, 'inception_v2_2016_08_28_frozen.mlmodel')
    mlmodel = tf_converter.convert(
        tf_model_path = tf_model_path,
        mlmodel_path = mlmodel_path,
        output_feature_names = ['InceptionV2/Predictions/Softmax:0'],
        input_name_shape_dict = {'input:0':[1,244,224,3]},
        image_input_names = ['input:0'],
        red_bias = -1,
        green_bias = -1,
        blue_bias = -1,
        image_scale = 2.0/255.0)

    #Test predictions on an image
    self._test_coreml_model_image_input(
        tf_model_path = tf_model_path,
        coreml_model = mlmodel,
        input_tensor_name = 'input:0',
        output_tensor_name = 'InceptionV2/Predictions/Softmax:0',
        img_size = 224)

  def test_googlenet_v4_slim(self):
    url = 'https://storage.googleapis.com/download.tensorflow.org/models/inception_v4_2016_09_09_frozen.pb.tar.gz'
    tf_model_dir = _download_file(url = url)
    tf_model_path = os.path.join(TMP_MODEL_DIR, 'inception_v4_2016_09_09_frozen.pb')

    mlmodel_path = os.path.join(TMP_MODEL_DIR, 'inception_v4_2016_09_09_frozen.mlmodel')
    mlmodel = tf_converter.convert(
        tf_model_path = tf_model_path,
        mlmodel_path = mlmodel_path,
        output_feature_names = ['InceptionV4/Logits/Predictions:0'],
        input_name_shape_dict = {'input:0':[1,299,299,3]},
        image_input_names = ['input:0'],
        red_bias = -1,
        green_bias = -1,
        blue_bias = -1,
        image_scale = 2.0/255.0)

    #Test predictions on an image
    self._test_coreml_model_image_input(
        tf_model_path = tf_model_path,
        coreml_model = mlmodel,
        input_tensor_name = 'input:0',
        output_tensor_name = 'InceptionV4/Logits/Predictions:0',
        img_size = 299)


  def test_mobilenet_v1_100_224(self):
    url = 'https://storage.googleapis.com/download.tensorflow.org/models/mobilenet_v1_1.0_224_frozen.tgz'
    tf_model_dir = _download_file(url = url)
    tf_model_path = os.path.join(TMP_MODEL_DIR, 'mobilenet_v1_1.0_224/frozen_graph.pb')

    mlmodel_path = os.path.join(TMP_MODEL_DIR, 'mobilenet_v1_1.0_224.mlmodel')
    mlmodel = tf_converter.convert(
        tf_model_path = tf_model_path,
        mlmodel_path = mlmodel_path,
        output_feature_names = ['MobilenetV1/Predictions/Softmax:0'],
        input_name_shape_dict = {'input:0':[1,224,224,3]},
        image_input_names = ['input:0'],
        red_bias = -1,
        green_bias = -1,
        blue_bias = -1,
        image_scale = 2.0/255.0)

    #Test predictions on an image
    self._test_coreml_model_image_input(
        tf_model_path = tf_model_path,
        coreml_model = mlmodel,
        input_tensor_name = 'input:0',
        output_tensor_name = 'MobilenetV1/Predictions/Softmax:0',
        img_size = 224)

  def test_mobilenet_v2_100_224(self):
    url = 'https://storage.googleapis.com/download.tensorflow.org/models/mobilenet_v1_1.0_224_frozen.tgz'
    tf_model_dir = _download_file(url = url)
    tf_model_path = os.path.join(TMP_MODEL_DIR, 'mobilenet_v1_1.0_224/frozen_graph.pb')

    mlmodel_path = os.path.join(TMP_MODEL_DIR, 'mobilenet_v1_1.0_224.mlmodel')
    mlmodel = tf_converter.convert(
        tf_model_path = tf_model_path,
        mlmodel_path = mlmodel_path,
        output_feature_names = ['MobilenetV1/Predictions/Softmax:0'],
        input_name_shape_dict = {'input:0':[1,224,224,3]},
        image_input_names = ['input:0'],
        red_bias = -1,
        green_bias = -1,
        blue_bias = -1,
        image_scale = 2.0/255.0)

    #Test predictions on an image
    self._test_coreml_model_image_input(
        tf_model_path = tf_model_path,
        coreml_model = mlmodel,
        input_tensor_name = 'input:0',
        output_tensor_name = 'MobilenetV1/Predictions/Softmax:0',
        img_size = 224)


  def test_mobilenet_v1_75_192(self):
    url = 'https://storage.googleapis.com/download.tensorflow.org/models/mobilenet_v1_0.75_192_frozen.tgz'
    tf_model_dir = _download_file(url = url)
    tf_model_path = os.path.join(TMP_MODEL_DIR, 'mobilenet_v1_0.75_192/frozen_graph.pb')

    mlmodel_path = os.path.join(TMP_MODEL_DIR, 'mobilenet_v1_0.75_192.mlmodel')
    mlmodel = tf_converter.convert(
        tf_model_path = tf_model_path,
        mlmodel_path = mlmodel_path,
        output_feature_names = ['MobilenetV1/Predictions/Softmax:0'],
        input_name_shape_dict = {'input:0':[1,192,192,3]},
        image_input_names = ['input:0'],
        red_bias = -1,
        green_bias = -1,
        blue_bias = -1,
        image_scale = 2.0/255.0)

    #Test predictions on an image
    self._test_coreml_model_image_input(
        tf_model_path = tf_model_path,
        coreml_model = mlmodel,
        input_tensor_name = 'input:0',
        output_tensor_name = 'MobilenetV1/Predictions/Softmax:0',
        img_size = 192)

  def test_mobilenet_v1_50_160(self):
    url = 'https://storage.googleapis.com/download.tensorflow.org/models/mobilenet_v1_0.50_160_frozen.tgz'
    tf_model_dir = _download_file(url = url)
    tf_model_path = os.path.join(TMP_MODEL_DIR, 'mobilenet_v1_0.50_160/frozen_graph.pb')

    mlmodel_path = os.path.join(TMP_MODEL_DIR, 'mobilenet_v1_0.50_160.mlmodel')
    mlmodel = tf_converter.convert(
        tf_model_path = tf_model_path,
        mlmodel_path = mlmodel_path,
        output_feature_names = ['MobilenetV1/Predictions/Softmax:0'],
        input_name_shape_dict = {'input:0':[1,160,160,3]},
        image_input_names = ['input:0'],
        red_bias = -1,
        green_bias = -1,
        blue_bias = -1,
        image_scale = 2.0/255.0)

    #Test predictions on an image
    self._test_coreml_model_image_input(
        tf_model_path = tf_model_path,
        coreml_model = mlmodel,
        input_tensor_name = 'input:0',
        output_tensor_name = 'MobilenetV1/Predictions/Softmax:0',
        img_size = 160)

  #@unittest.skip("Failing GPU backend: related to https://github.com/tf-coreml/tf-coreml/issues/26")
  def test_style_transfer(self):
    url = 'https://storage.googleapis.com/download.tensorflow.org/models/stylize_v1.zip'
    tf_model_dir = _download_file(url = url)
    tf_model_path = os.path.join(TMP_MODEL_DIR, 'stylize_quantized.pb')
    mlmodel_path = os.path.join(TMP_MODEL_DIR, 'stylize_quantized.mlmodel')
    # ? style transfer image size and style number?
    mlmodel = tf_converter.convert(
        tf_model_path = tf_model_path,
        mlmodel_path = mlmodel_path,
        output_feature_names = ['Squeeze:0'],
        input_name_shape_dict = {'input:0':[1,256,256,3], 'style_num:0':[26]})

    # Test predictions on an image
    input_tensors = [('input:0',[1,256,256,3]),
                     ('style_num:0',[26])]

    self.err_thresh = 0.5
    self._test_tf_model(
        tf_model_path = tf_model_path,
        coreml_model = mlmodel,
        input_tensors = input_tensors,
        output_tensor_names = ['Squeeze:0'],
        data_modes = ['image', 'onehot_0'],
        delta = 1e-2,
        use_cpu_only = True,
        scale = 1,
        bias = 0,
        img_size = 256,
        sequence_inputs = {'style_num:0'})

def _test_coreml_model_image_input(tf_model_path, coreml_model,
      input_tensor_name, output_tensor_name, img_size, useCPUOnly = False):
    """Test single image input conversions.
    tf_model_path - the TF model
    coreml_model - converted CoreML model
    input_tensor_name - the input image tensor name
    output_tensor_name - the output tensor name
    img_size - size of the image
    """

    img_np, img = _load_image(TEST_IMAGE ,resize_to=(img_size, img_size))
    img_tf = np.expand_dims(img_np, axis = 0)
    img_tf[:,:,:,0] = 2.0/255 * img_tf[:,:,:,0] - 1
    img_tf[:,:,:,1] = 2.0/255 * img_tf[:,:,:,1] - 1
    img_tf[:,:,:,2] = 2.0/255 * img_tf[:,:,:,2] - 1

    #evaluate the TF model
    tf.reset_default_graph()
    graph_def = graph_pb2.GraphDef()
    with open(tf_model_path, "rb") as f:
        graph_def.ParseFromString(f.read())
    g = tf.import_graph_def(graph_def)
    with tf.Session(graph=g) as sess:
      image_input_tensor = sess.graph.get_tensor_by_name('import/' + input_tensor_name)
      output = sess.graph.get_tensor_by_name('import/' + output_tensor_name)
      tf_out = sess.run(output,feed_dict={image_input_tensor: img_tf})
    if len(tf_out.shape) == 4:
      tf_out = np.transpose(tf_out, (0,3,1,2))
    tf_out_flatten = tf_out.flatten()

    #evaluate CoreML
    coreml_input_name = input_tensor_name.replace(':', '__').replace('/', '__')
    coreml_output_name = output_tensor_name.replace(':', '__').replace('/', '__')
    coreml_input = {coreml_input_name: img}

    #Test the default CoreML evaluation
    coreml_out = coreml_model.predict(coreml_input, useCPUOnly = useCPUOnly)[coreml_output_name]
    coreml_out_flatten = coreml_out.flatten()
    print (coreml_out_flatten)
    # compare_tf_coreml_outputs(tf_out_flatten, coreml_out_flatten)





if __name__=='__main__':
    # #Download model
    # url = 'https://storage.googleapis.com/download.tensorflow.org/models/inception_v3_2016_08_28_frozen.pb.tar.gz'
    # tf_model_dir = _download_file(url = url)
    # tf_model_path = os.path.join(TMP_MODEL_DIR, 'inception_v3_2016_08_28_frozen.pb')

    # #Convert to coreml
    # mlmodel_path = os.path.join(TMP_MODEL_DIR, 'inception_v3_2016_08_28.mlmodel')
    # mlmodel = tf_converter.convert(
    #     tf_model_path = tf_model_path,
    #     mlmodel_path = mlmodel_path,
    #     output_feature_names = ['InceptionV3/Predictions/Softmax:0'],
    #     input_name_shape_dict = {'input:0':[1,299,299,3]},
    #     image_input_names = ['input:0'],
    #     red_bias = -1,
    #     green_bias = -1,
    #     blue_bias = -1,
    #     image_scale = 2.0/255.0)

    # #Test predictions on an image
    # _test_coreml_model_image_input(
    #     tf_model_path = tf_model_path,
    #     coreml_model = mlmodel,
    #     input_tensor_name = 'input:0',
    #     output_tensor_name = 'InceptionV3/Predictions/Softmax:0',
    #     img_size = 299)

    #Download model
    url = 'https://storage.googleapis.com/download.tensorflow.org/models/inception5h.zip'
    tf_model_dir = _download_file(url = url)
    tf_model_path = os.path.join(TMP_MODEL_DIR, 'tensorflow_inception_graph.pb')

    #Convert to coreml
    mlmodel_path = os.path.join(TMP_MODEL_DIR, 'googlenet_v1_nonslim.mlmodel')
    mlmodel = tf_converter.convert(
        tf_model_path = tf_model_path,
        mlmodel_path = mlmodel_path,
        output_feature_names = ['softmax2:0'],
        input_name_shape_dict = {'input:0':[1,224,224,3]},
        image_input_names = ['input:0'],
        red_bias = -1,
        green_bias = -1,
        blue_bias = -1,
        image_scale = 2.0/255.0)

    #Test predictions on an image
    _test_coreml_model_image_input(
        tf_model_path = tf_model_path,
        coreml_model = mlmodel,
        input_tensor_name = 'input:0',
        output_tensor_name = 'softmax2:0',
        img_size = 224)


    print("convert ok!")



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.utils import *
from mmdnn.conversion.common.IR.graph_pb2 import TensorShape, AttrValue
from mmdnn.conversion.common.DataStructure.graph import Graph, GraphNode


def load_protobuf_from_file(container, filename):
    with open(filename, 'rb') as fin:
        file_content = fin.read()

    # First try to read it as a binary file.
    try:
        container.ParseFromString(file_content)
        print("Parse file [%s] with binary format successfully." % (filename))
        return container

    except Exception as e:  # pylint: disable=broad-except
        print ("Info: Trying to parse file [%s] with binary format but failed with error [%s]." % (filename, str(e)))

    # Next try to read it as a text file.
    try:
        from google.protobuf import text_format
        text_format.Parse(file_content.decode('UTF-8'), container, allow_unknown_extension=True)
        print("Parse file [%s] with text format successfully." % (filename))
    except text_format.ParseError as e:
        raise IOError("Cannot parse file %s: %s." % (filename, str(e)))

    return container


class IRGraphNode(GraphNode):

    @staticmethod
    def replace_scope(name):
        return name.replace('/', '_')

    @property
    def IR_layer(self):
        return self.layer

    @property
    def name(self):
        return self.layer.name

    @property
    def type(self):
        return self.layer.op

    def set_attrs(self, attrs):
        assign_IRnode_values(self.layer, attrs)


    def get_attr(self, name, default_value = None):
        if name in self.layer.attr:
            attr = self.layer.attr[name]
            field = attr.WhichOneof('value')
            val = getattr(attr, field) if field else default_value
            if not val:
                return val
            if isinstance(val, AttrValue.ListValue):
                if val.ListFields():
                    return list(val.ListFields()[0][1])
                else:
                    return val.ListFields()
            else:
                return val.decode('utf-8') if isinstance(val, bytes) else val
        else:
            return default_value


class IRGraph(Graph):

    @staticmethod
    def shapeToStr(tensor_shape, keep_minus_one = False):
        ret = ""
        first = True
        for e in tensor_shape.dim:
            if e.size != -1 or keep_minus_one:
                if first == False:
                    ret += ", "
                ret += str(e.size)
                first = False
        return ret


    def __init__(self, filename):
        model = graph_pb2.GraphDef()
        load_protobuf_from_file(model, filename)
        super(IRGraph, self).__init__(model)


    def filter_node(self):
        self.layer_map = dict(filter(lambda layer: layer[1].in_edges or layer[1].out_edges, self.layer_map.items()))


    def build(self):
        for layer in self.model.node:
            self.layer_map[layer.name] = IRGraphNode(layer)
            self.layer_name_map[layer.name] = layer.name

        for i, layer in enumerate(self.model.node):
            for pred in layer.input:
                self._make_connection(pred, layer.name)

        self.filter_node()
        super(IRGraph, self).build()
        self.input_layers = list(filter(lambda x: self.layer_map[x].type != 'Constant', self.input_layers))


    def rebuild(self):
        self.input_layers.clear()
        self.output_layers.clear()
        self.topological_sort.clear()
        self.filter_node()
        super(IRGraph, self).build()
        self.input_layers = list(filter(lambda x: self.layer_map[x].type != 'Constant', self.input_layers))


    def clear_out_scope_node(self):

        def _clear_list_out_scope(list_):
            for idx in range(len(list_) -1, -1, -1):
                node = self.get_node(list_[idx])
                if node.type != 'Scope' and node.get_attr('scope'):
                    del list_[idx]

        _clear_list_out_scope(self.input_layers)
        _clear_list_out_scope(self.topological_sort)
        _clear_list_out_scope(self.output_layers)




import tensorflow as tf


def save_model(MainModel, network_filepath, weight_filepath, dump_filepath, dump_tag = 'SERVING'):
    if dump_tag == 'SERVING':
        tag_list = [tf.saved_model.tag_constants.SERVING]
    else:
        tag_list = [tf.saved_model.tag_constants.TRAINING]
    input, model = MainModel.KitModel(weight_filepath)
    with tf.Session() as sess:
        sess.run(tf.global_variables_initializer())

        builder = tf.saved_model.builder.SavedModelBuilder(dump_filepath)

        tensor_info_input = tf.saved_model.utils.build_tensor_info(input)
        tensor_info_output = tf.saved_model.utils.build_tensor_info(model)

        prediction_signature = (
            tf.saved_model.signature_def_utils.build_signature_def(
                inputs={'input': tensor_info_input},
                outputs={'output': tensor_info_output},
                method_name=tf.saved_model.signature_constants.PREDICT_METHOD_NAME
            )
        )

        builder.add_meta_graph_and_variables(
            sess,
            tag_list,
            signature_def_map={
                tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY: prediction_signature
            }
        )

        save_path = builder.save()

    print('Tensorflow file is saved as [{}], generated by [{}.py] and [{}].'.format(
        save_path, network_filepath, weight_filepath))



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import os

import math
import mxnet as mx
import numpy as np
from mmdnn.conversion.common.IR.IR_graph import IRGraph, IRGraphNode
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.IR.graph_pb2 import NodeDef, GraphDef, DataType
from mmdnn.conversion.common.DataStructure.emitter import Emitter
from mmdnn.conversion.common.utils import *
from mmdnn.conversion.rewriter.folder import Folder

class MXNetEmitter(Emitter):

    dtype_map = {
        graph_pb2.DT_FLOAT16    : "float16",
        graph_pb2.DT_FLOAT32    : "float32",
        graph_pb2.DT_FLOAT64    : "float64",
        graph_pb2.DT_INT32      : "int32",
        graph_pb2.DT_UINT8      : "uint8"
    }

    activation_map = {
        "relu"      : "Relu",
        "sigmoid"   : "Sigmoid",
        "tanh"      : "Tanh",
        "elu"       : "Elu"
    }

    transpose_map = {
        1 : 2,
        2 : 3,
       -1 : 1
    }

    naive_scope_pattern = []

    channels_last = ['NDHWC', 'NHWC']

    def __init__(self, model):
        super(MXNetEmitter, self).__init__()
        from six import string_types as _string_types

        if isinstance(model, _string_types):
            network_path = model
            self.weight_loaded = False
        elif len(model) == 3:
            network_path = model[0]
            weight_path = model[1]
            self.output_weights_file = model[2]
            self.weights = np.load(weight_path).item()
            self.weight_loaded = True
            self.output_weights = dict()
        else:
            raise ValueError("the # of input arguments [{}] is not supported" % len(model))

        self.IR_graph = IRGraph(network_path)
        self.IR_graph.build()

        folder = Folder(self.IR_graph, self.weights)
        folder.fold()

    @property
    def header_code(self):
        return """import mxnet as mx
import numpy as np
import math

# mxnet-cpu only support channel first, default convert the model and weight as channel first

def RefactorModel():
"""


    def gen_code(self, phase):
        self.IR_layer_map = dict()
        self.add_body(0, self.header_code)
        for layer in self.IR_graph.topological_sort:
            self.IR_layer_map[layer] = self.IR_graph.get_node(layer)

        shape = dict()
        for layer in self.IR_graph.topological_sort:
            current_node = self.IR_graph.get_node(layer)
            node_type = current_node.type


            if len(current_node.in_edges) == 0:
                current_node.in_edges.append('data')

            if node_type.lower() in MXNetEmitter.activation_map:
                func = getattr(self, "emit_Activation")
                line = func(current_node, MXNetEmitter.activation_map[node_type.lower()].lower())
                self.add_body(1, line)

            elif hasattr(self, "emit_" + node_type):
                func = getattr(self, "emit_" + node_type)
                line = func(current_node)
                if line != None:
                    self.add_body(1, line)
            else:
                print("MXNet Emitter has not supported operator [%s]." % (node_type))
                self.emit_UNKNOWN(current_node)

            if node_type == "DataInput":
                cur_shape = list()
                first = True
                for dim in current_node.IR_layer.attr["shape"].shape.dim:
                    if dim.size == -1 and first:
                        cur_shape.append(1)
                        print("Detect input layer [{}] using infer batch size, set it as default value [1]".format(current_node.name))
                    else:
                        if dim.size == -1:
                            print("Warning: user should change input size manually")
                        cur_shape.append(dim.size)
                    first = False

                cur_shape.insert(1, cur_shape.pop())
                shape[current_node.name] = ', '.join('%s' % i for i in cur_shape)
                self.input_name_shape = {current_node.name: tuple(cur_shape)}


        if self.weight_loaded:
            fullpath = os.path.abspath(self.output_weights_file)
            dirname = os.path.dirname(fullpath)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            with open(self.output_weights_file, 'wb') as outfile:
                np.save(outfile, self.output_weights)

        comment = "\n    # if a GPU is available, change mx.cpu() to mx.gpu()"
        # We use the real_name for specifying the input layer in data_names
        # since MXNet API wants the actual name of the layer. On the other
        # hand, the module API wants the last symbol in the symbol chain, so
        # for the output node we need to use the actual python variable name
        # of the last layer (real_variable_name).
        last_line = "{:<15} = mx.mod.Module(symbol = {}, context = mx.cpu(), data_names = ['{}'])".format(
            "model",
            ', '.join([self.IR_graph.get_node(name).real_variable_name for name in self.IR_graph.output_layers if self.IR_graph.get_node(name).type !='Pack' and self.IR_graph.get_node(name).type != 'Shape']),
            ', '.join([self.IR_graph.get_node(name).real_name for name in self.IR_graph.input_layers if self.IR_graph.get_node(name).type != 'Const']))

        self.add_body(1, comment)
        self.add_body(1, last_line)
        self.add_body(1, "return model")


        self.add_body(0, "")
        for code in self.layers_codes.values():
            self.add_body(0, code)

        weight_code = ""
        if not self.weight_loaded:
            weight_code += "# emitter does not detect any import weights, you may generate weights file manually\n"

        weight_code += self.gen_weight_code(shape, phase)

        main_code = "if __name__ == '__main__':\n    model = RefactorModel()\n"
        if self.weight_loaded:
            main_code += "    # remember to adjust params path\n    model = deploy_weight(model, '{}')\n".format(self.output_weights_file)

        if phase == 'train':
            train_code = """def train(model):
    import logging
    logging.getLogger().setLevel(logging.DEBUG)
    model.fit(train_iter, # train data
            eval_data = val_iter, # validation data
            optimizer = 'sgd', # Defaults to 'sgd'
            optimizer_params = {'learning_rate':0.01}, # use fixed learning rate
            eval_metric = 'acc', # report accuracy during training, other possible predefined metrics are: 'ce', 'f1', 'mae', 'mse', 'rmse', 'top_k_accuracy'
            batch_end_callback = mx.callback.Speedometer(batch_size, 100), # output progress for each 100 data batches
            num_epoch = 10) # train for at most 10 dataset passes\n\n
"""
            code = self.body_code + weight_code + train_code + main_code
        else:
            test_code = """from collections import namedtuple
Batch = namedtuple('Batch', ['data'])


def get_image(url, show=False):
    import cv2
    # download and show the image
    fname = mx.test_utils.download(url)
    img = cv2.cvtColor(cv2.imread(fname), cv2.COLOR_BGR2RGB)
    if img is None:
        return None
    if show:
        import matplotlib.pyplot as plt
        plt.imshow(img)
        plt.axis('off')
    # convert into format (batch, RGB, width, height)
    img = cv2.resize(img, (224, 224))
    img = np.swapaxes(img, 0, 2)
    img = np.swapaxes(img, 1, 2)
    img = img[np.newaxis, :]
    return img


def predict(model, labels, url):
    # to show the image, change the argument show into True
    img = get_image(url, show = False)
    # compute the predict probabilities
    model.forward(Batch([mx.nd.array(img)]))
    prob = model.get_outputs()[0].asnumpy()
    # print the top-5
    prob = np.squeeze(prob)
    a = np.argsort(prob)[::-1]
    for i in a[0:5]:
        print('prbability = %f, class = %s' %(prob[i], labels[i]))\n\n
"""

            main_code += """
    # # call function predict
    # with open('synset.txt', 'r') as f:
    #     labels = [l.rstrip() for l in f]
    # predict(model, labels, 'http://writm.com/wp-content/uploads/2016/08/Cat-hd-wallpapers.jpg')
"""

            code = self.body_code + weight_code + test_code + main_code

        return code


    def gen_weight_code(self, shape, phase):
        str = "def deploy_weight(model, weight_file):\n"
        str += """
    if weight_file == None:
        return

    try:
        weights_dict = np.load(weight_file).item()
    except:
        weights_dict = np.load(weight_file, encoding='bytes').item()

    arg_params = dict()
    aux_params = dict()
    for weight_name, weight_data in weights_dict.items():
        weight_name = str(weight_name)
        if "moving" in weight_name:
            aux_params[weight_name] = mx.nd.array(weight_data)
        else:
            arg_params[weight_name] = mx.nd.array(weight_data)

"""
        if phase == 'train':
            str += "    model.bind(for_training = True, data_shapes = ["
        else:
            str += "    model.bind(for_training = False, data_shapes = ["
        first = True
        for k, v in shape.items():
            if not first:
                str += ", "
            str += "('" + k + "', " + "(" + v + "))"
            first = False
        str += "])\n"
        str += "    model.set_params(arg_params = arg_params, aux_params = aux_params, allow_missing = True, allow_extra=True)\n\n    return model\n\n\n"
        return str


    @staticmethod
    def calculate_same_pad(data_shape, kernel, stride):
        if (data_shape % stride == 0):
            pad = max(kernel - stride, 0)
        else:
            pad = max(kernel - (data_shape % stride), 0)
        if pad % 2 == 0:
            return False, pad
        else:
            return True, pad


    @staticmethod
    def transfer_pad(pad_list):
        defuse_pad = False
        pad = list()

        assert len(pad_list) % 2 == 0
        mid = int(len(pad_list)/2)
        pad_first = pad_list[1:mid-1]
        pad_second = pad_list[mid+1:-1]

        for i in range(0, mid-2):
            if not pad_first[i] == pad_second[i]:
                defuse_pad = True

        if defuse_pad:
            pad.extend([0] * 4)
            for i in range(0, mid-2):
                pad.extend([pad_first[i], pad_second[i]])
        else:
            pad = pad_first

        return defuse_pad, pad


    @staticmethod
    def transpose(data, dim):
        if dim == 1:
            data = data.transpose((2, 1, 0))
        elif dim == 2:
            data = data.transpose((3, 2, 0, 1))
        elif dim == 3:
            data = data.transpose((4, 3, 0, 1, 2))
        else:
            raise ValueError("The weight of dim {} cannot transpose" % dim)

        return data


    def set_pad(self, IR_node, code, pad, _max_pool):
        if _max_pool:
            constant_value = "float('-inf')"
        else:
            constant_value = "0.0"

        code = "{:<15} = mx.sym.pad(data = {}, mode = 'constant', pad_width={}, constant_value = {}, name = '{}')".format(
                IR_node.variable_name + "_pad",
                self.parent_variable_name(IR_node),
                tuple(pad),
                constant_value,
                IR_node.name + "_pad")

        for e in IR_node.in_edges:
            e = e.split(':')[0]
            if e == 'data':
                continue
            self.IR_layer_map[e].out_edges = [x if not self.IR_layer_map[x.split(':')[0]].name == IR_node.variable_name else IR_node.variable_name + "_pad" for x in self.IR_layer_map[e].out_edges]

        return code


    def emit_UNKNOWN(self, IR_node):
        print(IR_node.name)


    def emit_FullyConnected(self, IR_node):
        if self.weight_loaded:
            weight_dict = self.weights[IR_node.name]
            parent = self.IR_graph.get_parent(IR_node.name, [0])
            while parent.type == "Flatten" or parent.type == 'Dropout':
                parent = self.IR_graph.get_parent(parent.name, [0])
            dim = len(parent.layer.attr['_output_shapes'].list.shape[0].dim)
            if dim > 2:
                original_dims = weight_dict['weights'].shape
                dims = [i.size for i in parent.layer.attr['_output_shapes'].list.shape[0].dim[1:]] + [-1]
                weight_dict['weights'] = np.reshape(weight_dict['weights'], dims)
                weight_dict['weights'] = np.transpose(weight_dict['weights'], [dim - 2] + list(range(0, dim - 2)) + [dim - 1])
                weight_dict['weights'] = np.reshape(weight_dict['weights'], original_dims)
            self.output_weights[IR_node.name + "_weight"] = weight_dict['weights'].transpose((1, 0))

        num_hidden = IR_node.IR_layer.attr["units"].i
        no_bias = not IR_node.IR_layer.attr["use_bias"].b
        if not no_bias and self.weight_loaded:
            self.output_weights[IR_node.name + "_bias"] = weight_dict['bias']

        code = "{:<15} = mx.sym.FullyConnected(data = {}, num_hidden = {}, no_bias = {}, name = '{}')".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                num_hidden,
                no_bias,
                IR_node.name)

        return code


    def _emit_convolution(self, IR_node, pattern):
        if self.weight_loaded:
            weight_dict = self.weights[IR_node.name]
            weights = weight_dict['weights']

        dim = len(IR_node.IR_layer.attr["kernel_shape"].list.i) - 2

        kernel = list()
        for idx in range(0, dim):
            kernel.append(IR_node.IR_layer.attr["kernel_shape"].list.i[idx])

        stride = list()
        for e in IR_node.IR_layer.attr["strides"].list.i[1:-1]:
            stride.append(e)

        dilate = list()
        for e in IR_node.IR_layer.attr["dilations"].list.i[1:-1]:
            dilate.append(e)
        if dilate == []: dilate = [1, 1]
        dilate = ', '.join('%s' % i for i in dilate)

        defuse_pad = False
        pad = list()
        if "pads" in IR_node.IR_layer.attr:
            output_shape = list()
            for e in IR_node.IR_layer.attr["_output_shapes"].list.shape[0].dim:
                output_shape.append(e.size)

            # print("Warning: MXNet Convolution Layer pad does not match IR Convolution Layer pad")
            defuse_pad, pad = MXNetEmitter.transfer_pad(IR_node.IR_layer.attr["pads"].list.i)

        num_filter = 0
        if pattern == "Deconvolution":
            num_filter = IR_node.IR_layer.attr["kernel_shape"].list.i[-2]
        else:
            num_filter = IR_node.IR_layer.attr["kernel_shape"].list.i[-1]

        use_bias = IR_node.get_attr('use_bias', False)
        if use_bias and self.weight_loaded:
            self.output_weights[IR_node.name + "_bias"] = weight_dict['bias']

        if pattern == "DepthwiseConv":
            num_group = IR_node.IR_layer.attr["kernel_shape"].list.i[-2]
            num_filter = num_filter * num_group
            pattern = "Convolution"
            if self.weight_loaded:
                weights = np.swapaxes(weights, -1, -2)

        else:
            num_group = IR_node.get_attr('group', 1)

        # layout = IR_node.IR_layer.attr["data_format"].s
        if dim == 1:
            layout = 'NCW'
        elif dim == 2:
            layout = 'NCHW'
        elif dim == 3:
            layout = 'NCDHW'

        if self.weight_loaded:
            # if layout not in MXNetEmitter.channels_last:
            weights = MXNetEmitter.transpose(weights, dim)
            self.output_weights[IR_node.name + "_weight"] = weights

        code = ""
        if not defuse_pad:
            code += "{:<15} = mx.sym.{}(data={}, kernel={}, stride={}, dilate = ({}), pad={}, num_filter = {}, num_group = {}, no_bias = {}, layout = '{}', name = '{}')".format(
                IR_node.variable_name,
                pattern,
                self.parent_variable_name(IR_node),
                tuple(kernel),
                tuple(stride),
                dilate,
                tuple(pad),
                num_filter,
                num_group,
                not use_bias,
                layout,
                IR_node.name)
        else:
            code += self.set_pad(IR_node, code, pad, False)
            code += "\n    {:<15} = mx.sym.{}(data={}, kernel={}, stride={}, dilate = ({}), num_filter = {}, num_group = {}, no_bias = {}, layout = '{}', name = '{}')".format(
                IR_node.variable_name,
                pattern,
                IR_node.variable_name + "_pad",
                tuple(kernel),
                tuple(stride),
                dilate,
                num_filter,
                num_group,
                not use_bias,
                layout,
                IR_node.name)

        return code


    def emit_Conv(self, IR_node):
        return self._emit_convolution(IR_node, "Convolution")


    def emit_DepthwiseConv(self, IR_node):
        return self._emit_convolution(IR_node, "DepthwiseConv")


    def emit_ConvTranspose(self, IR_node):
        return self._emit_convolution(IR_node, "Deconvolution")


    def emit_DataInput(self, IR_node):
        shape = list()
        shape.extend(IR_node.IR_layer.attr["shape"].list.i)

        code = "{:<15} = mx.sym.var('{}')".format(IR_node.variable_name, IR_node.name)
        return code


    # Add LeakyReLU Elu(slope not support)
    def emit_Activation(self, IR_node, act_type):

        act_type = act_type
        func_name = ""

        if act_type == "elu":
            func_name = "LeakyReLU"
        else:
            func_name = "Activation"

        code = "{:<15} = mx.sym.{}(data = {}, act_type = '{}', name = '{}')".format(
                IR_node.variable_name,
                func_name,
                self.parent_variable_name(IR_node),
                act_type,
                IR_node.name)

        return code


    def emit_BatchNorm(self, IR_node):
        IR_node_after = self.IR_graph.get_son(IR_node.name, [0])
        if IR_node_after.type == 'Scale':
            if self.weight_loaded:
                weight_dict = self.weights[IR_node.name]
                weight_dict_scale = self.weights[IR_node_after.name]

            # axis = IR_node.IR_layer.attr["axis"].i
            axis = 1
            eps = IR_node.IR_layer.attr["epsilon"].f
            momentum = IR_node.IR_layer.attr["momentum"].f

            fix_gamma = not IR_node.IR_layer.attr["scale"].b

            if self.weight_loaded:
                if not fix_gamma:
                #     self.output_weights[IR_node.name + "_gamma"] = np.multiply(weight_dict['scale'], weight_dict_scale['scale'])
                # self.output_weights[IR_node.name + "_beta"] = np.multiply(weight_dict['bias'], weight_dict_scale['scale']) + weight_dict_scale['bias']
                    self.output_weights[IR_node.name + "_gamma"] = weight_dict['scale']
                self.output_weights[IR_node.name + "_beta"] = weight_dict['bias']

            # not supported yet
            use_global_stats = "False"
            if self.weight_loaded:
                self.output_weights[IR_node.name + "_moving_var"] = weight_dict['var']
                self.output_weights[IR_node.name + "_moving_mean"] = weight_dict['mean']

            code = "{:<15} = mx.sym.BatchNorm(data = {}, axis = {}, eps = {}, momentum = {}, fix_gamma = {}, use_global_stats = {}, name = '{}')".format(
                    IR_node.variable_name,
                    self.parent_variable_name(IR_node),
                    axis,
                    eps,
                    momentum,
                    fix_gamma,
                    use_global_stats,
                    IR_node.name)

            return code

        else:
            if self.weight_loaded:
                weight_dict = self.weights[IR_node.name]

            # axis = IR_node.IR_layer.attr["axis"].i
            axis = 1
            eps = IR_node.IR_layer.attr["epsilon"].f
            momentum = IR_node.IR_layer.attr["momentum"].f

            fix_gamma = not IR_node.IR_layer.attr["scale"].b

            if self.weight_loaded:
                if not fix_gamma:
                    self.output_weights[IR_node.name + "_gamma"] = weight_dict['scale']
                self.output_weights[IR_node.name + "_beta"] = weight_dict['bias']

            # not supported yet
            use_global_stats = "False"
            if self.weight_loaded:
                self.output_weights[IR_node.name + "_moving_var"] = weight_dict['var']
                self.output_weights[IR_node.name + "_moving_mean"] = weight_dict['mean']

            code = "{:<15} = mx.sym.BatchNorm(data = {}, axis = {}, eps = {}, momentum = {}, fix_gamma = {}, use_global_stats = {}, name = '{}')".format(
                    IR_node.variable_name,
                    self.parent_variable_name(IR_node),
                    axis,
                    eps,
                    momentum,
                    fix_gamma,
                    use_global_stats,
                    IR_node.name)

            return code

    def emit_Scale(self, IR_node):
        if self.weight_loaded:
            weight_dict = self.weights[IR_node.name]

        # axis = IR_node.IR_layer.attr["axis"].i
        axis = 1
        eps = 0.0
        momentum = 0.0

        fix_gamma = not IR_node.IR_layer.attr["scale"].b

        if self.weight_loaded:
            if not fix_gamma:
                self.output_weights[IR_node.name + "_gamma"] = weight_dict['scale']
            self.output_weights[IR_node.name + "_beta"] = weight_dict['bias']

        # not supported yet
        use_global_stats = "False"
        if self.weight_loaded:
            self.output_weights[IR_node.name + "_moving_var"] = weight_dict['scale_var']
            self.output_weights[IR_node.name + "_moving_mean"] = weight_dict['scale_mean']

        code = "{:<15} = mx.sym.BatchNorm(data = {}, axis = {}, eps = {}, momentum = {}, fix_gamma = {}, use_global_stats = {}, name = '{}')".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                axis,
                eps,
                momentum,
                fix_gamma,
                use_global_stats,
                IR_node.name)

        return code



    def emit_Pool(self, IR_node):

        global_pool = IR_node.IR_layer.attr["global_pooling"].b

        kernel = list()
        if global_pool:
            kernel = [1] * (len(IR_node.IR_layer.attr["strides"].list.i) - 2)
        else:
            for e in IR_node.IR_layer.attr["kernel_shape"].list.i[1:-1]:
                kernel.append(e)

        pool_type = IR_node.get_attr('pooling_type').lower()

        stride = list()
        for e in IR_node.IR_layer.attr["strides"].list.i[1:-1]:
            stride.append(e)

        defuse_pad = False
        pad = list()
        if "pads" in IR_node.IR_layer.attr:
            output_shape = list()
            for e in IR_node.IR_layer.attr["_output_shapes"].list.shape[0].dim:
                output_shape.append(e.size)

            # print("Warning: MXNet Pooling Layer pad does not match IR Pooling Layer pad")
            defuse_pad, pad = MXNetEmitter.transfer_pad(IR_node.IR_layer.attr["pads"].list.i)
        code = ""
        if not defuse_pad:
            code += "{:<15} = mx.sym.Pooling(data = {}, global_pool = {}, kernel={}, pool_type = '{}', stride={}, pad={}, name = '{}')".format(
                    IR_node.variable_name,
                    self.parent_variable_name(IR_node),
                    global_pool,
                    tuple(kernel),
                    pool_type,
                    tuple(stride),
                    tuple(pad),
                    IR_node.name)
        else:
            code += self.set_pad(IR_node, code, pad, pool_type == "max")
            code += "\n    {:<15} = mx.sym.Pooling(data = {}, global_pool = {}, kernel={}, pool_type = '{}', stride={}, name = '{}')".format(
                    IR_node.variable_name,
                    IR_node.variable_name + "_pad",
                    global_pool,
                    tuple(kernel),
                    pool_type,
                    tuple(stride),
                    IR_node.name)

        return code


    def emit_SoftmaxOutput(self, IR_node):

        code = "{:<15} = mx.sym.SoftmaxOutput(data = {}, name = 'softmax')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node)
        )

        return code


    def emit_Softmax(self, IR_node):

        code = ""

        if len(IR_node.out_edges) == 0:
            code = "{:<15} = mx.sym.SoftmaxOutput(data = {}, name = 'softmax')".format(
                    IR_node.variable_name,
                    self.parent_variable_name(IR_node))
        else:
            axis = IR_node.IR_layer.attr["dim"].i
            code = "{:<15} = mx.sym.softmax(data = {}, axis = {}, name = '{}')".format(
                    IR_node.variable_name,
                    self.parent_variable_name(IR_node),
                    axis,
                    IR_node.name)

        return code


    def emit_Squeeze(self, IR_node):
        return self.emit_Flatten(IR_node)


    # def emit_ConvTranspose(self, IR_node):
    #     if self.weight_loaded:
    #         weight_dict = self.weights[IR_node.name]
    #         weights = weight_dict['weights']

    #     dim = len(IR_node.IR_layer.attr["kernel_shape"].list.i) - 2

    #     kernel = list()
    #     for idx in range(0, dim):
    #         kernel.append(IR_node.IR_layer.attr["kernel_shape"].list.i[idx])

    #     stride = list()
    #     for e in IR_node.IR_layer.attr["strides"].list.i[1:-1]:
    #         stride.append(e)

    #     dilate = list()
    #     for e in IR_node.IR_layer.attr["dilations"].list.i[1:-1]:
    #         dilate.append(e)
    #     dilate = ', '.join('%s' % i for i in dilate)

    #     defuse_pad = False
    #     pad = list()
    #     if "pads" in IR_node.IR_layer.attr:
    #         output_shape = list()
    #         for e in IR_node.IR_layer.attr["_output_shapes"].list.shape[0].dim:
    #             output_shape.append(e.size)

    #         # print("Warning: MXNet Deconvolution Layer pad does not match IR Deconvolution Layer pad")
    #         defuse_pad, pad = MXNetEmitter.transfer_pad(IR_node.IR_layer.attr["pads"].list.i)
    #     pad = ', '.join('%s' % i for i in pad)

    #     kernel = ', '.join('%s' % i for i in kernel)
    #     stride = ', '.join('%s' % i for i in stride)

    #     num_filter = IR_node.IR_layer.attr["kernel_shape"].list.i[-2]
    #     no_bias = not IR_node.IR_layer.attr["use_bias"].b
    #     if not no_bias and self.weight_loaded:
    #         self.output_weights[IR_node.replace_scope(IR_node.name) + "_bias"] = weight_dict['bias']

    #     # layout = IR_node.IR_layer.attr["data_format"].s
    #     if dim == 1:
    #         layout = 'NCW'
    #     elif dim == 2:
    #         layout = 'NCHW'
    #     elif dim == 3:
    #         layout = 'NCDHW'

    #     if self.weight_loaded:
    #         # if layout not in MXNetEmitter.channels_last:
    #         weights = MXNetEmitter.transpose(weights, dim)
    #         self.output_weights[IR_node.replace_scope(IR_node.name) + "_weight"] = weights

    #     code = ""
    #     if not defuse_pad:
    #         code = "{:<15} = mx.sym.Deconvolution(data = {}, kernel = ({}), stride = ({}), dilate = ({}), pad = ({}), num_filter = {}, no_bias = {}, layout = '{}', name = '{}')".format(
    #                 IR_node.replace_scope(IR_node.name),
    #                 IR_node.replace_scope(IR_node.in_edges[0]),
    #                 kernel,
    #                 stride,
    #                 dilate,
    #                 pad,
    #                 num_filter,
    #                 no_bias,
    #                 layout,
    #                 IR_node.replace_scope(IR_node.name))
    #     else:
    #         code = self.set_pad(IR_node, code, pad)
    #         code += "\n    {:<15} = mx.sym.Deconvolution(data = {}, kernel = ({}), stride = ({}), dilate = ({}), num_filter = {}, no_bias = {}, layout = '{}', name = '{}')".format(
    #                 IR_node.replace_scope(IR_node.name), IR_node.replace_scope(IR_node.name) + "_pad", kernel, stride, dilate, num_filter, no_bias, layout, IR_node.replace_scope(IR_node.name))

    #     return code


    def emit_Embedding(self, IR_node):

        input_dim = IR_node.IR_layer.attr["input_dim"].i
        output_dim = IR_node.IR_layer.attr["output_dim"].i
        dtype = MXNetEmitter.dtype_map.get(IR_node.layer.attr["dtype"].type, "float32")

        weight_dict = self.weights[IR_node.name]

        if self.weight_loaded:
            self.output_weights[IR_node.name + "_weight"] = weight_dict['weights']

        code = "{:<15} = mx.sym.Embedding(data = {}, input_dim = {}, output_dim = {}, dtype = '{}', name = '{}')".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                input_dim,
                output_dim,
                dtype,
                IR_node.name)

        return code


    def emit_LeakyRelu(self, IR_node):
        alpha = IR_node.IR_layer.attr['alpha'].f
        code = "{:<15} = mx.sym.LeakyReLU(data = {}, slope = {}, name = '{}')".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                alpha,
                IR_node.name
        )
        return code

    def emit_PRelu(self, IR_node):
        slope = IR_node.get_attr('gamma')
        code = "{:<15} = mx.sym.LeakyReLU(data = {}, slope = {}, act_type = '{}', name = '{}')".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                slope,
                'prelu',
                IR_node.name
        )
        return code

    def emit_Elu(self, IR_node):
        alpha = IR_node.IR_layer.attr['alpha'].f
        code = "{:<15} = mx.sym.LeakyReLU(data = {}, slope = {}, act_type = {}, name = '{}')".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                alpha,
                'elu',
                IR_node.name
        )
        return code

    def emit_Dropout(self, IR_node):
        p = IR_node.IR_layer.attr["keep_prob"].f
        mode = IR_node.IR_layer.attr["mode"].s.lower().decode() if 'mode' in IR_node.layer.attr else 'training'
        code = "{:<15} = mx.sym.Dropout(data = {}, p = {}, mode = '{}', name = '{}')".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                p,
                mode,
                IR_node.name)

        return code


    # reverse cannot support yet
    def emit_Reshape(self, IR_node):
        shape = list()
        for e in IR_node.IR_layer.attr["shape"].list.i:
            shape.append(e)
        shape = ', '.join('%s' % i for i in shape)
        reverse = False

        code = "{:<15} = mx.sym.reshape(data = {}, shape = ({}), reverse = {}, name = '{}')".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                shape,
                reverse,
                IR_node.name)

        return code


    def emit_Flatten(self, IR_node):
        # code = "{:<15} = mx.sym.transpose(data = {}, axes = (0, 2, 3, 1))\n".format("trans", self.parent_variable_name(IR_node))
        code = "{:<15} = mx.sym.flatten(data = {}, name = '{}')".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                IR_node.name)

        return code


    @staticmethod
    def _convert_axis(IR_node, axis):
        ndim = len(IR_node.layer.attr['_output_shapes'].list.shape[0].dim)
        if axis == 0:
            return 0
        elif axis == ndim - 1:
            return 1
        else:
            return axis + 1


    def emit_Concat(self, IR_node):
        dim = MXNetEmitter._convert_axis(IR_node, IR_node.IR_layer.attr["axis"].i)
        code = "{:<15} = mx.sym.concat({}, dim = {}, name = '{}')".format(
                IR_node.variable_name,
                ', '.join(self.parent_variable_name(IR_node, [idx]) for idx in range(len(IR_node.in_edges))),
                dim,
                IR_node.name)

        return code


    def emit_Cast(self, IR_node):
        dtype = IR_node.IR_layer.attr["dtype"].type
        code = "{:<15} = mx.sym.cast(data = {}, dtype = {}, name = '{}')".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                dtype,
                IR_node.name)

        return code


    def emit_Expand_dims(self, IR_node):
        axis = IR_node.IR_layer.attr["axis"].i
        code = "{:<15} = mx.sym.expand_dims(data = {}, axis = {}, name = '{}')".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                axis,
                IR_node.name)

        return code


    def emit_Pad(self, IR_node):
        mode = IR_node.IR_layer.attr["mode"].s.lower().decode()
        pad_width = list()
        pad_width.extend([0]*4)
        padding = convert_onnx_pad_to_tf(IR_node.get_attr("pads"))[1:-1]
        for padding_pair in padding:
            pad_width.extend(padding_pair)

        pad_width = ', '.join('%s' % i for i in pad_width)

        code = "{:<15} = mx.sym.pad(data = {}, mode = '{}', pad_width = ({}), name = '{}')".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                mode,
                pad_width,
                IR_node.name)

        return code


    def emit_Add(self, IR_node):
        code = "{:<15} = mx.sym.broadcast_add({}, {})".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                self.parent_variable_name(IR_node, [1]))

        return code


    def emit_Mul(self, IR_node):

        code = "{:<15} = mx.sym.broadcast_mul({}, {})".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                self.parent_variable_name(IR_node, [1]))

        return code


    def emit_ReduceMean(self, IR_node):
        axes = IR_node.layer.attr['axes'].list.i[:]
        axes = ','.join('%s' % MXNetEmitter.transpose_map[i] for i in axes)

        code = "{:<15} = mx.sym.mean(data = {}, axis = ({}), keepdims = {})".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                axes,
                IR_node.layer.attr['keepdims'].b)

        return code


    def emit_LRN(self, IR_node):
        code = "{:<15} = mx.sym.LRN(data = {}, alpha = {}, beta = {}, knorm = {}, nsize = {}, name = '{}')".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                IR_node.layer.attr['alpha'].f,
                IR_node.layer.attr['beta'].f,
                IR_node.layer.attr['k'].f,
                IR_node.layer.attr['size'].i * 2 - 1,
                IR_node.name)

        return code

    def emit_Constant(self, IR_node):
        # save the constant into weight dict
        if IR_node.get_attr('value'):
            value = IR_node.get_attr('value')
        else:
            value = self.weights[IR_node.name]['value']
    
        if not isinstance(value, list):
            self.output_weights[IR_node.name + '_weight'] = [value] # mxnet's bug, it does not surpport scalar weight. 
            code = "{:<15} = mx.sym.var(name = '{}', shape=(1,))".format(IR_node.variable_name, IR_node.name+'_weight')
        else:
            shape = np.array(value).shape
            self.output_weights[IR_node.name + '_weight'] = value

            code = "{:<15} = mx.sym.var(name = '{}', shape={})".format(IR_node.variable_name, IR_node.name+'_weight', shape)

        return code

    def emit_Sub(self, IR_node):
        code = "{:<15} = mx.sym.broadcast_sub({}, {})".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                self.parent_variable_name(IR_node, [1]))

        return code


    def emit_Relu6(self, IR_node):
        codes = list()
        codes.append(self.emit_Activation(IR_node, 'relu'))
        old_name = IR_node.variable_name
        IR_node.real_name = IR_node.real_name + "_clip"
        codes.append("{:<15} = mx.sym.clip({}, a_min=0, a_max=6, name='{}')".format(
            IR_node.real_variable_name,
            old_name,
            IR_node.real_name))

        return codes


    def emit_Slice(self, IR_node):

        starts = IR_node.get_attr('starts')
        starts = [starts[0], starts[-1]] + starts[1:-1]
        ends = IR_node.get_attr('ends')
        ends = [ends[0], ends[-1]] + ends[1:-1]
        ends = [i if i else None for i in ends]
        strides = IR_node.get_attr('strides')
        if strides:
            strides = [strides[0], strides[-1]] + strides[1:-1]

        code =  "{:<15} = mx.sym.slice({}, begin={}, end={}, step={}, name='{}')".format(
            IR_node.real_variable_name,
            self.parent_variable_name(IR_node),
            starts,
            ends,
            strides,
            IR_node.name
        )
        return code

    def emit_Const(self, IR_node):
        pass

    def emit_Shape(self, IR_node):
        code = "{:<15} = mx.sym.var(init = mx.init.Constant({}.infer_shape({}={})[1][0]), name='{}')".format(
            IR_node.real_variable_name,
            self.parent_variable_name(IR_node),
            list(self.input_name_shape.keys())[0],
            list(self.input_name_shape.values())[0],
            IR_node.name
        )
        return code

    def emit_Pack(self, IR_node):
        pass

    def emit_Unsqueeze(self, IR_node):
        axis = IR_node.get_attr('axes')[0]
        code = "{:<15} = mx.sym.expand_dims(data = {}, axis = {}, name = '{}')".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                axis,
                IR_node.name)

        return code

    def emit_Unstack(self, IR_node):
        squeeze_axis = axis = IR_node.get_attr('axis')
        num = IR_node.get_attr('num')
        if num is None:
            args_str = ""
            for input_name in self.IR_graph.input_layers:
                if self.IR_graph.get_node(input_name).type!='Const':
                    args_str += '{}={}, '.format(self.IR_graph.get_node(input_name).real_variable_name, self.data_input_shape[input_name])

            args_str = args_str[:-2]
            num_outputs = "{}.infer_shape({})[1][0][{}]".format(
                IR_node.variable_name,
                args_str,
                axis
            )
        else:
            num_outputs = num

        code = "{:<15} = mx.sym.split({}, num_outputs={}, axis={}, squeeze_axis={})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            num_outputs,
            axis,
            squeeze_axis
        )
        return code

    def emit_Fill(self, IR_node):
        value = IR_node.get_attr('value')
        code = "{:<15} = mx.sym.full({}, {})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            value
        )
        return code

    def emit_Split(self, IR_node):
        axis = IR_node.get_attr('axis')
        num_outputs = IR_node.get_attr('split')

        if isinstance(num_outputs, list):
            raise NotImplementedError()
        code = "{:<15} = mx.sym.split({}, num_outputs={}, axis={})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            num_outputs,
            axis)

        return code


    def emit_Sigmoid(self, IR_node):
        code = "{:<15} = mx.sym.sigmoid(data={}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.name
        )
        return code


    def emit_Tanh(self, IR_node):
        code = "{:<15} = mx.sym.tanh(data={}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.name
        )
        return code


    def emit_Maxmum(self, IR_node):
        code = "{:<15} = mx.sym.maxmum({}, {}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            self.parent_variable_name(IR_node, [1]),
            IR_node.name
        )
        return code


    def emit_Minimum(self, IR_node):
        code = "{:<15} = mx.sym.minimum({}, {}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            self.parent_variable_name(IR_node, [1]),
            IR_node.name
        )
        return code


    def emit_Scope(self, IR_node):
        import re
        pattern = IR_node.pattern
        
        if pattern not in self.naive_scope_pattern and re.sub(r'(_\d+)*$', '', IR_node.pattern) not in self.naive_scope_pattern:
            origi_pattern = re.sub(r'(_\d+)*$', '', IR_node.pattern)
            func = getattr(self, "_emit_" + origi_pattern)
            code = func(IR_node)
        else:
            code = "{:<15} = __{}({})".format(
                IR_node.real_variable_name,
                IR_node.pattern,
                ', '.join(self.parent_variable_name(IR_node, s) for s in IR_node.in_edges))
            self._gen_scope_code(IR_node)
        return code


    def _gen_scope_code(self, scope_node):

        def _get_weight_related_op_name(node):
            weight_related_ops = ['Constant', 'Conv', 'FullyConnected', 'BatchNorm']
            op_type = node.type
            if op_type in weight_related_ops:
                return op_type, node.name

        def _scope_func(params, code, return_var):
            code = """
    def __call__(self, {}):
{}
        return {}
    """.format(params, code, ', '.join(return_var))
            return code

        class_inits = dict()

        body_code = str()
        for node_name in scope_node.topology_list:
            node = self.IR_graph.get_node(node_name)
            node_type = node.type

            if hasattr(self, "emit_" + node_type):
                func = getattr(self, "emit_" + node_type)
                line = func(node)
                if line != None:
                    body_code += "        " + line + '\n'
                    inits = _get_weight_related_op_name(node)
                    if inits:
                        if class_inits.get(inits[0], None):
                            class_inits[inits[0]].append(inits[1])
                        else:
                            class_inits[inits[0]] = list([inits[1]])
            else:
                print("MXNetEmitter has not supported operator [%s]." % (node_type))
                self.emit_UNKNOWN(node)

        # param_code does not need parameter slice.
        param_code = ', '.join('%s'  %self.IR_graph.get_node(s).real_variable_name for s in scope_node.in_edges)
        function_code = _scope_func(param_code, body_code, scope_node.return_variables)

        return class_inits, function_code


    def _emit_gru_cell(self, IR_node):
        if not self.layers_codes.get(IR_node.pattern, None):
            class_inits, func_code = self._gen_scope_code(IR_node)
            variables, variable_codes, init_code, func_code = self.process_inits_func_code(class_inits, func_code)

            states = [self.IR_graph.get_node(s).real_variable_name for s in IR_node.in_edges]
            states.pop(0)
            states_code = ', '.join(states)

            class_code ='''
class _{}(mx.rnn.BaseRNNCell):
    def __init__(self, {}):

{}

{}

            '''.format(IR_node.pattern,
            ', '.join(variables),
            init_code,
            func_code)
            self.layers_codes[IR_node.pattern] = class_code

            if not hasattr(self, 'pattern_variables'):
                self.pattern_variables = {IR_node.pattern: variables}
            else:
                self.pattern_variables[IR_node.pattern] = variables

            code = variable_codes
            code.append("{:<15} = _{}({})({})".format(
                IR_node.real_variable_name,
                IR_node.pattern,
                ', '.join(variables),
                ', '.join(self.parent_variable_name(IR_node, s) for s in IR_node.in_edges)))
        else:
            code = "{:<15} = _{}({})({})".format(
                IR_node.real_variable_name,
                IR_node.pattern,
                ', '.join(self.pattern_variables[IR_node.pattern]),
                ', '.join(self.parent_variable_name(IR_node, s) for s in IR_node.in_edges))

        return code


    def _emit_h_zero(self, IR_node):
        code = "{:<15} = mx.sym.full((1, {}), {})".format(
            IR_node.variable_name,
            IR_node.get_attr('fill_size'),
            IR_node.get_attr('fill_value')
        )
        return code
    

    def _emit_lstm_cell(self, IR_node):

        if not self.layers_codes.get(IR_node.pattern, None):
            class_inits, func_code = self._gen_scope_code(IR_node)
            variables, variable_codes, init_code, func_code = self.process_inits_func_code(class_inits, func_code)

            states = [self.IR_graph.get_node(s).real_variable_name for s in IR_node.in_edges]
            states.pop(0)
            states_code = ', '.join(states)

            class_code ='''
class _{}(mx.rnn.BaseRNNCell):
    def __init__(self, {}):

{}

{}

            '''.format(IR_node.pattern,
            ', '.join(variables),
            init_code,
            func_code)
            self.layers_codes[IR_node.pattern] = class_code

            if not hasattr(self, 'pattern_variables'):
                self.pattern_variables = {IR_node.pattern: variables}
            else:
                self.pattern_variables[IR_node.pattern] = variables

            code = variable_codes
            code.append("{:<15} = _{}({})({})".format(
                IR_node.real_variable_name,
                IR_node.pattern,
                ', '.join(variables),
                ', '.join(self.parent_variable_name(IR_node, s) for s in IR_node.in_edges)))
        else:
            code = "{:<15} = _{}({})({})".format(
                IR_node.real_variable_name,
                IR_node.pattern,
                ', '.join(self.pattern_variables[IR_node.pattern]),
                ', '.join(self.parent_variable_name(IR_node, s) for s in IR_node.in_edges))

        return code


    def process_inits_func_code(self, class_inits, func_code):
        init_code = str()
        variables = list()
        variable_codes = list()
        for k, v in class_inits.items():
            if k == 'FullyConnected':
                for i, name in enumerate(class_inits[k]):
                    variable_name = self.IR_graph.get_node(name).variable_name
                    variables.append("W_" + variable_name)
                    variable_codes.append("W_{:<15} = mx.sym.var(name='{}_weight')".format(variable_name, name))
                    init_code += "        self.W_{} = W_{}\n".format(variable_name, variable_name)

                    if self.weight_loaded and self.weights[name].get('bias', None).any() != None:
                        variable_codes.append("B_{:<15} = mx.sym.var(name='{}_bias')".format(variable_name, name))
                        variables.append("B_" + variable_name)
                        init_code += "        self.B_{} = B_{}\n".format(variable_name, variable_name)
                        func_code = func_code.replace("name = '{}'".format(name), "name = '{}', weight = self.W_{}, bias = self.B_{}".format(name, variable_name, variable_name))
                    else:
                        func_code = func_code.replace("name = '{}'".format(name), "name = '{}', weight = self.W_{}".format(name, variable_name))
            elif k == 'Constant':
                for name in class_inits[k]:
                    variable_name = self.IR_graph.get_node(name.replace('_weight', '')).variable_name
                    variables.append(variable_name)
                    constant_line = self.emit_Constant(self.IR_graph.get_node(name.replace('_weight', '')))
                    variable_codes.append("{:<15} = {}".format(variable_name, '='.join(constant_line.split('=')[1:])))
                    init_code += "        self.{} = {}\n".format(variable_name, variable_name)
                    func_code = func_code.replace(constant_line, constant_line.split('=')[0] + ' = self.'+constant_line.split('=')[0])
            else:
                raise NotImplementedError

        return variables, variable_codes, init_code, func_code




import sys


class ConversionError(Exception):
    '''
    an abtract class
    '''
    pass


def print_stderr(msg):
    '''
    a function to print information to the std
    '''
    sys.stderr.write('%s\n' % msg)



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
from collections import OrderedDict
from mmdnn.conversion.common.DataStructure.graph import GraphNode, Graph
# from tensorflow.core.framework.node_def_pb2 import NodeDef
# from tensorflow.core.framework import attr_value_pb2


class DarknetGraphNode(GraphNode):

    def __init__(self, layer):

        super(DarknetGraphNode, self).__init__(layer)


    @property
    def name(self):
        return self.layer['name']


    @property
    def type(self):
        return self.layer['type']


    @property
    def dk_layer(self):
        return self.layer


    def get_attr(self, name, default_value = None):
        if name in self.layer['attr'].keys():
            return self.layer['attr'][name]
        else:
            return default_value


class DarknetGraph(Graph):

    def __init__(self, model):
        # pass

        super(DarknetGraph, self).__init__(model)
        self.layer_num_map = {}
        self.model = model
        self.weights = {}
        self.original_list = OrderedDict()

    @staticmethod
    def dim_str_to_int(input_dim):
        if type(input_dim) == list:
            return [int(i) for i in input_dim]

    @staticmethod
    def conv_output_width(width, padding, kernel_size, stride):
        return (width + 2*padding - kernel_size)/stride + 1

    @staticmethod
    def conv_output_height(height, padding, kernel_size, stride):
        return (height + 2*padding - kernel_size)/stride + 1

    def build(self):

        for i, block in enumerate(self.model):
            # print("\n")
            # print(i)
            # print(block)

            # continue
            node = OrderedDict()
            if block['type'] == 'net':
                node['name'] = 'dk_Input'
                node['input'] = ['data']
                node['type'] = 'DataInput'
                node['input_dim'] = ['-1']
                # NHWC
                node['input_dim'].append(block['height'])
                node['input_dim'].append(block['width'])
                node['input_dim'].append(block['channels'])
                input_param = OrderedDict()
                input_param['shape'] = self.dim_str_to_int(node['input_dim'])
                input_param['_output_shape'] = self.dim_str_to_int(node['input_dim'])
                node['attr'] = input_param
                self.layer_map[node['name']] = DarknetGraphNode(node)
                self.original_list[node['name']] = DarknetGraphNode(node)
                self.layer_num_map[i] = node['name']
                pre_node_name = node['name']

            elif block['type'] == 'convolutional':
                conv_layer = OrderedDict()
                conv_layer['input'] = [pre_node_name]

                input_shape = self.layer_map[pre_node_name].get_attr('_output_shape')
                w = input_shape[1]
                h = input_shape[2]
                channels = input_shape[3]
                # assert False

                if 'name' in block.keys():
                    conv_layer['name'] = block['name']
                else:
                    conv_layer['name'] = 'layer%d-conv' % i
                conv_layer['type'] = 'Conv'

                convolution_param = OrderedDict()
                convolution_param['num_output'] = int(block['filters'])
                convolution_param['kernel_size'] = int(block['size'])
                convolution_param['kernel'] = [int(block['size']), int(block['size']), channels, int(block['filters'])]
                convolution_param['pad'] = int(block['pad'])

                if block['pad'] == '1':
                    convolution_param['padding'] = int(convolution_param['kernel_size'])//2
                convolution_param['stride'] = int(block['stride'])
                if block['batch_normalize'] == '1':
                    convolution_param['bias_term'] = 'false'
                else:
                    convolution_param['bias_term'] = 'true'
                output_w = self.conv_output_width(w ,convolution_param['padding'], convolution_param['kernel_size'], convolution_param['stride'])
                output_h = self.conv_output_height(h ,convolution_param['padding'], convolution_param['kernel_size'], convolution_param['stride'])
                convolution_param['_output_shape'] = [-1, output_w, output_h, convolution_param['num_output']]
                conv_layer['attr'] = convolution_param
                self.layer_map[conv_layer['name']] = DarknetGraphNode(conv_layer)
                self.original_list[conv_layer['name']] = DarknetGraphNode(conv_layer)
                pre_node_name = conv_layer['name']

                if block['batch_normalize'] == '1':
                    bn_layer = OrderedDict()
                    bn_layer['input'] = [pre_node_name]


                    input_shape = self.layer_map[pre_node_name].get_attr('_output_shape')
                    if 'name' in block.keys():
                        bn_layer['name'] = '%s-bn' % block['name']
                    else:
                        bn_layer['name'] = 'layer%d-bn' % i
                    bn_layer['type'] = 'BatchNorm'
                    batch_norm_param = OrderedDict()
                    batch_norm_param['use_global_stats'] = True
                    batch_norm_param['_output_shape'] = convolution_param['_output_shape']
                    batch_norm_param['bias_term'] = True
                    batch_norm_param['scale'] = True
                    bn_layer['attr'] = batch_norm_param


                    self.layer_map[bn_layer['name']] = DarknetGraphNode(bn_layer)
                    self.original_list[bn_layer['name']] = DarknetGraphNode(bn_layer)

                    pre_node_name = bn_layer['name']


                if block['activation'] != 'linear':
                    relu_layer = OrderedDict()
                    relu_layer['input'] = [pre_node_name]
                    if 'name' in block.keys():
                        relu_layer['name'] = '%s-act' % block['name']
                    else:
                        relu_layer['name'] = 'layer%d-act' % i
                    relu_layer['type'] = 'ReLU'
                    relu_param = OrderedDict()
                    if block['activation'] == 'leaky':
                        relu_layer['type'] = 'leakyReLU'
                        relu_param['negative_slope'] = '0.1'
                    relu_param['_output_shape'] = input_shape
                    relu_layer['attr'] = relu_param
                    self.layer_map[relu_layer['name']] = DarknetGraphNode(relu_layer)
                    self.layer_num_map[i] = relu_layer['name']
                    self.original_list[relu_layer['name']] = DarknetGraphNode(relu_layer)
                    pre_node_name = relu_layer['name']

                else:
                    self.layer_num_map[i] = bn_layer['name']


            elif block['type'] == 'maxpool':
                max_layer = OrderedDict()
                max_layer['input'] = [pre_node_name]
                if 'name' in block.keys():
                    max_layer['name'] = block['name']
                else:
                    max_layer['name'] = 'layer%d-maxpool' % i
                max_layer['type'] = 'Pooling'
                pooling_param = OrderedDict()
                pooling_param['kernel_size'] = int(block['size'])
                pooling_param['stride'] = int(block['stride'])
                pooling_param['pool'] = 'MAX'
                pooling_param['padding'] = 0
                if 'pad' in block.keys() and int(block['pad']) == 1:
                    pooling_param['padding'] = (int(block['size'])-1)/2

                input_shape = self.layer_map[pre_node_name].get_attr('_output_shape')
                w = input_shape[1]
                h = input_shape[2]
                output_w = (w + 2*pooling_param['padding'])/pooling_param['stride']
                output_h = (h + 2*pooling_param['padding'])/pooling_param['stride']

                pooling_param['_output_shape'] = [-1, output_w, output_h, input_shape[-1]]
                max_layer['attr'] = pooling_param
                self.layer_map[max_layer['name']] = DarknetGraphNode(max_layer)
                self.original_list[max_layer['name']] = DarknetGraphNode(max_layer)
                self.layer_num_map[i] = max_layer['name']
                pre_node_name = max_layer['name']

            elif block['type'] == 'avgpool':
                avg_layer = OrderedDict()

                avg_layer['input'] = [pre_node_name]
                if 'name' in block.keys():
                    avg_layer['name'] = block['name']
                else:
                    avg_layer['name'] = 'layer%d-avgpool' % i
                avg_layer['type'] = 'Pooling'
                pooling_param = OrderedDict()
                input_shape = self.layer_map[pre_node_name].get_attr('_output_shape')
                pooling_param['_output_shape'] = [-1, 1, 1, input_shape[-1]]
                pooling_param['pool'] = 'AVG'
                avg_layer['attr'] = pooling_param
                self.layer_map[avg_layer['name']] = DarknetGraphNode(avg_layer)
                self.original_list[avg_layer['name']] = DarknetGraphNode(avg_layer)
                self.layer_num_map[i] = avg_layer['name']
                pre_node_name = avg_layer['name']

            elif block['type'] == 'route':
                prev = block['layers'].split(',') #[-1,61]
                if len(prev) == 1:
                    prev_layer_id = i + int(prev[0])
                    self.layer_num_map[i] = self.layer_num_map[prev_layer_id]
                    pre_node_name = self.layer_num_map[i]
                elif len(prev) == 2:
                    input_list = []
                    input_shape = []
                    route_layer = OrderedDict()
                    for p in prev:
                        if int(p)>0:

                            input_name = self.layer_num_map[int(p)+1]
                            input_list.append(input_name)
                            input_shape.append(self.layer_map[input_name].get_attr('_output_shape'))

                        else:
                            prev_layer_id = i + int(p)
                            input_name = self.layer_num_map[prev_layer_id]
                            input_shape.append(self.layer_map[input_name].get_attr('_output_shape'))
                            input_list.append(input_name)
                    route_param = OrderedDict()


                    shape_ = 0
                    for shape in input_shape:
                        shape_ += shape[-1]
                    route_param['axis'] = 3
                    route_param['_output_shape'] = input_shape[0][:-1] + [shape_]
                    route_layer['input'] = input_list

                    if 'name' in block.keys():
                        route_layer['name'] = block['name']
                    else:
                        route_layer['name'] = 'layer%d-concat' % i

                    route_layer['type'] = 'Concat'
                    route_layer['attr'] = route_param

                    self.layer_map[route_layer['name']] = DarknetGraphNode(route_layer)
                    self.original_list[route_layer['name']] = DarknetGraphNode(route_layer)
                    self.layer_num_map[i] = route_layer['name']
                    pre_node_name = route_layer['name']

            elif block['type'] == 'shortcut':
                prev_layer_id1 = i + int(block['from'])
                prev_layer_id2 = i - 1
                bottom1 = self.layer_num_map[prev_layer_id1]
                bottom2 = self.layer_num_map[prev_layer_id2]
                input_shape = self.layer_map[bottom2].get_attr('_output_shape')
                shortcut_layer = OrderedDict()
                shortcut_layer['input'] = [bottom1, bottom2]
                # print(shortcut_layer['input'] )
                if 'name' in block.keys():
                    shortcut_layer['name'] = block['name']
                else:
                    shortcut_layer['name'] = 'layer%d-shortcut' % i
                shortcut_layer['type'] = 'Add'
                eltwise_param = OrderedDict()
                eltwise_param['operation'] = 'SUM'
                eltwise_param['_output_shape'] = input_shape
                shortcut_layer['attr'] = eltwise_param


                self.layer_map[shortcut_layer['name']] = DarknetGraphNode(shortcut_layer)
                self.original_list[shortcut_layer['name']] = DarknetGraphNode(shortcut_layer)
                self.layer_num_map[i] = shortcut_layer['name']
                pre_node_name = shortcut_layer['name']

                if block['activation'] != 'linear':
                    relu_layer = OrderedDict()
                    relu_layer['input'] = [pre_node_name]
                    if 'name' in block.keys():
                        relu_layer['name'] = '%s-act' % block['name']
                    else:
                        relu_layer['name'] = 'layer%d-act' % i
                    relu_layer['type'] = 'ReLU'
                    relu_param = OrderedDict()
                    relu_param['_output_shape'] = input_shape
                    if block['activation'] == 'leaky':

                        relu_param['negative_slope'] = '0.1'

                    relu_layer['attr'] = relu_param
                    self.layer_map[relu_layer['name']] = DarknetGraphNode(relu_layer)
                    self.original_list[relu_layer['name']] = DarknetGraphNode(relu_layer)
                    pre_node_name = relu_layer['name']

            elif block['type'] == 'connected':
                fc_layer = OrderedDict()
                fc_layer['input'] = [pre_node_name]
                if 'name' in block.keys():
                    fc_layer['name'] = block['name']
                else:
                    fc_layer['name'] = 'layer%d-fc' % i
                fc_layer['type'] = 'InnerProduct'
                fc_param = OrderedDict()
                fc_param['num_output'] = int(block['output'])
                input_shape = self.layer_map[pre_node_name].get_attr('_output_shape')
                fc_param['_output_shape'] = input_shape[:-1] + [fc_param['num_output']]
                fc_layer['attr'] = fc_param
                self.layer_map[fc_layer['name']] = DarknetGraphNode(fc_layer)
                self.original_list[fc_layer['name']] = DarknetGraphNode(fc_layer)
                self.layer_num_map[i] = fc_layer['name']
                pre_node_name = fc_layer['name']

                if block['activation'] != 'linear':
                    relu_layer = OrderedDict()
                    relu_layer['input'] = [pre_node_name]
                    if 'name' in block.keys():
                        relu_layer['name'] = '%s-act' % block['name']
                    else:
                        relu_layer['name'] = 'layer%d-act' % i
                    relu_layer['type'] = 'ReLU'
                    relu_param = OrderedDict()
                    if block['activation'] == 'leaky':

                        relu_param['negative_slope'] = '0.1'
                    relu_param['_output_shape'] = fc_param['_output_shape']
                    relu_layer['attr'] = relu_param
                    self.layer_map[relu_layer['name']] = DarknetGraphNode(relu_layer)
                    self.original_list[relu_layer['name']] = DarknetGraphNode(relu_layer)
                    pre_node_name = relu_layer['name']

            elif block['type'] == 'softmax':
                sm_layer = OrderedDict()

                sm_layer['input'] = [pre_node_name]
                if 'name' in block.keys():
                    sm_layer['name'] = block['name']
                else:
                    sm_layer['name'] = 'layer%d-softmax' % i
                sm_layer['type'] = 'Softmax'
                softmax_param = OrderedDict()
                input_shape = self.layer_map[pre_node_name].get_attr('_output_shape')
                softmax_param['_output_shape'] = input_shape
                sm_layer['attr'] = softmax_param
                self.layer_map[sm_layer['name']] = DarknetGraphNode(sm_layer)
                self.original_list[sm_layer['name']] = DarknetGraphNode(sm_layer)
                self.layer_num_map[i] = sm_layer['name']
                pre_node_name = sm_layer['name']

            elif block['type'] == 'yolo':

                yolo_layer = OrderedDict()
                yolo_layer['input'] = [pre_node_name]
                if 'name' in block.keys():
                    yolo_layer['name'] = block['name']
                else:
                    yolo_layer['name'] = 'layer%d-yolo' % i
                yolo_layer['type'] = 'yolo'
                yolo_param = OrderedDict()
                yolo_param['truth_thresh'] = float(block['truth_thresh'])
                yolo_param['random'] = float(block['random'])
                yolo_param['ignore_thresh'] = float(block['ignore_thresh'])
                yolo_param['jitter'] = float(block['jitter'])
                yolo_param['num'] = int(block['num'])
                yolo_param['classes'] = int(block['classes'])
                anchors = [int(t) for t in block['anchors'].split(',')]
                yolo_param['anchors'] = anchors
                mask = [int(t) for t in block['mask'].split(',')]
                yolo_param['mask'] = mask

                yolo_layer['attr'] = yolo_param
                self.layer_map[yolo_layer['name']] = DarknetGraphNode(yolo_layer)
                self.original_list[yolo_layer['name']] = DarknetGraphNode(yolo_layer)
                self.layer_num_map[i] = yolo_layer['name']

            elif block['type'] == 'upsample':

                input_shape = self.layer_map[pre_node_name].get_attr('_output_shape')
                upsample_layer = OrderedDict()
                upsample_layer['input'] = [pre_node_name]
                if 'name' in block.keys():
                    upsample_layer['name'] = block['name']
                else:
                    upsample_layer['name'] = 'layer%d-upsample' % i
                upsample_layer['type'] = 'upsample'
                upsample_param = OrderedDict()
                stride = block['stride']
                upsample_param['scales'] = [int(stride), int(stride)]
                upsample_param['_output_shape'] = [input_shape[0]] + [q*int(stride) for q in input_shape[1:3]] + [input_shape[-1]]
                upsample_layer['attr'] = upsample_param
                self.layer_map[upsample_layer['name']] = DarknetGraphNode(upsample_layer)
                self.original_list[upsample_layer['name']] = DarknetGraphNode(upsample_layer)
                self.layer_num_map[i] = upsample_layer['name']
                pre_node_name = upsample_layer['name']

            elif block['type'] == 'cost':
                continue

            # spacetodepth
            elif block['type'] == 'reorg':
                input_shape = self.layer_map[pre_node_name].get_attr('_output_shape')
                reorg_layer = OrderedDict()
                reorg_layer['input'] = [pre_node_name]
                if 'name' in block.keys():
                    reorg_layer['name'] = block['name']
                else:
                    reorg_layer['name'] = 'layer%d-reorg' % i

                reorg_layer['type'] = 'SpaceToDepth'
                reorg_param = OrderedDict()
                stride = int(block['stride'])
                reorg_param['strides'] = stride
                reorg_param['_output_shape'] = [-1, input_shape[1]/stride, input_shape[2]/stride, input_shape[3]*stride*stride]
                reorg_layer['attr'] = reorg_param

                self.layer_map[reorg_layer['name']] = DarknetGraphNode(reorg_layer)
                self.original_list[reorg_layer['name']] = DarknetGraphNode(reorg_layer)
                self.layer_num_map[i] = reorg_layer['name']
                pre_node_name = reorg_layer['name']


            elif block['type'] == 'region':
                # print(block)
                region_layer = OrderedDict()
                region_layer['input'] = [pre_node_name]
                if 'name' in block.keys():
                    region_layer['name'] = block['name']
                else:
                    region_layer['name'] = 'layer%d-region' % i
                region_layer['type'] = 'region'
                region_param = OrderedDict()
                region_param['softmax'] = int(block['softmax'])
                region_param['thresh'] = float(block['thresh'])
                region_param['random'] = float(block['random'])
                region_param['jitter'] = float(block['jitter'])
                region_param['num'] = int(block['num'])
                region_param['classes'] = int(block['classes'])
                region_param['coords'] = int(block['coords'])
                region_param['rescore'] = int(block['rescore'])
                region_param['object_scale'] = int(block['object_scale'])

                region_param['noobject_scale'] = int(block['noobject_scale'])
                region_param['class_scale'] = int(block['class_scale'])
                region_param['coord_scale'] = int(block['coord_scale'])

                region_param['bias_match'] = int(block['bias_match'])
                region_param['absolute'] = int(block['absolute'])

                anchors = [float(t) for t in block['anchors'].split(',')]
                region_param['anchors'] = anchors

                region_layer['attr'] = region_param
                # print(region_layer)
                self.layer_map[region_layer['name']] = DarknetGraphNode(region_layer)
                self.original_list[region_layer['name']] = DarknetGraphNode(region_layer)
                self.layer_num_map[i] = region_layer['name']
                # assert False


            else:
                print('unknown layer type %s ' % block['type'])
                print(block,"\n")
                assert False



        for layer in self.layer_map:
            for pred in self.layer_map[layer].layer['input']:
                if pred not in self.layer_map.keys() and pred != 'data':
                    print(pred)
                    print("::::::::::::: unknown input :::::::::::::")
                    assert False

                self._make_connection(pred, layer)

        super(DarknetGraph, self).build()




from mmdnn.conversion.rewriter.rewriter import UnitRewriterBase
import numpy as np
import re

class GRURewriter(UnitRewriterBase):

    def __init__(self, graph, weights_dict):
        return super(GRURewriter, self).__init__(graph, weights_dict)
    
    def process_gru_cell(self, match_result):
        if 'gru_cell' not in match_result._pattern_to_op.keys():
            return
        kwargs = dict()
        top_node = match_result._pattern_to_op[match_result._name_to_pattern['gru_cell']]

        w_e = match_result.get_op("cell_kernel")
        w = self._weights_dict[w_e.name.replace('/read', '')]

        num_units = w.shape[1]//2
        input_size = w.shape[0] - num_units

        kwargs['num_units'] = num_units
        kwargs['input_size'] = input_size

        if hasattr(top_node, 'kwargs'):
            top_node.kwargs.update(kwargs)
        else:
            top_node.kwargs = kwargs


    def process_rnn_h_zero(self, match_result):
        if 'h_zero' not in match_result._name_to_pattern.keys():
            return
        kwargs = dict()
        top_node = match_result._pattern_to_op[match_result._name_to_pattern['h_zero']]

        fill_size = match_result.get_op('fill_size')
        fill_value = match_result.get_op('fill_value')

        kwargs['fill_size'] = fill_size.get_attr('value').int_val[0]
        kwargs['fill_value'] = fill_value.get_attr('value').float_val[0]

        if hasattr(top_node, 'kwargs'):
            top_node.kwargs.update(kwargs)
        else:
            top_node.kwargs = kwargs


    def process_match_result(self, match_result, pattern_name):
        if pattern_name == 'gru_cell':
            self.process_gru_cell(match_result)
        elif pattern_name == 'h_zero':
            if self.check_match_scope(match_result, 'GRUCellZeroState'):
                self.process_rnn_h_zero(match_result)

    '''For some short pattern, to avoid match other pattern, check it's scope'''
    def check_match_scope(self, match_result, scope_name):
        ops = match_result._pattern_to_op.values()

        for op in ops:
            op_name_splits = op.name.split('/')
            if len(op_name_splits) < 2:
                return False
            if re.sub(r'(_\d+)*$', '', op_name_splits[-2]) != scope_name:
                if len(op_name_splits) > 2:
                    if re.sub(r'(_\d+)*$', '', op_name_splits[-3]) != scope_name:
                        return False
                else:
                    return False
        return True


    def run(self):
        return super(GRURewriter, self).run(['gru_cell', 'h_zero'], 'tensorflow')


#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import os
import numpy as np
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.IR.graph_pb2 import NodeDef, GraphDef, DataType
from mmdnn.conversion.common.utils import *
from mmdnn.conversion.common.DataStructure.parser import Parser
from mmdnn.conversion.pytorch.pytorch_graph import PytorchGraph
import torch
import torchvision

class PytorchParser(Parser):

    layer_map = {
    'onnx::Conv': 'Conv',
    'onnx::Flatten': 'Flatten',
    'onnx::Gemm': 'FullyConnected',
    'onnx::MaxPool': 'Maxpool',
    'onnx::AveragePool': 'Avgpool',
    'onnx::Dropout': 'Dropout',
    'onnx::BatchNormalization': 'BatchNormalization',
    'onnx::Add': 'Add',
    'onnx::Concat': 'Concat',
    'onnx::Relu': 'Relu',
    'onnx::Tanh': 'Tanh',
    'onnx::Sigmoid': 'Sigmoid',
    'onnx::Mul': 'Mul'


    # TODO
    # 'max_pool2d': convert_maxpool,
    # 'onnx::Mul': convert_elementwise_mul,
    # 'onnx::Sub': convert_elementwise_sub,
    # 'onnx::ConvTranspose': convert_convtranspose,
    # 'onnx::LeakyRelu': convert_lrelu,
    # 'onnx::Sigmoid': convert_sigmoid,
    # 'onnx::Softmax': convert_softmax,
    # 'onnx::Selu': convert_selu,
    # 'onnx::Transpose': convert_transpose,
    # 'onnx::Reshape': convert_reshape,
    # 'onnx::MatMul': convert_matmul,
    # 'onnx::Gather': convert_gather,
    # 'onnx::ReduceSum': convert_reduce_sum,
    # 'onnx::Constant': convert_constant,
    # 'onnx::Upsample': convert_upsample,
    # 'onnx::Pad': convert_padding,
}


    ############
    # property #
    ############

    @property
    def src_graph(self):
        return self.pytorch_graph


    ####################
    # Public Functions #
    ####################

    def __init__(self, model_file_name, input_shape):
        super(PytorchParser, self).__init__()
        if not os.path.exists(model_file_name):
            print("Pytorch model file [{}] is not found.".format(model_file_name))
            assert False
        # test

        # cpu: https://github.com/pytorch/pytorch/issues/5286
        try:
            model = torch.load(model_file_name)
        except:
            model = torch.load(model_file_name, map_location='cpu')

        self.weight_loaded = True

        # Build network graph
        self.pytorch_graph = PytorchGraph(model)
        self.input_shape = tuple([1] + input_shape)
        self.pytorch_graph.build(self.input_shape)
        self.state_dict = self.pytorch_graph.state_dict
        self.shape_dict = self.pytorch_graph.shape_dict


    def gen_IR(self):

        for layer in self.src_graph.topological_sort:
            current_node = self.src_graph.get_node(layer)
            onnx_node_type = current_node.type
            node_type = PytorchParser.layer_map[onnx_node_type]


            if hasattr(self, "rename_" + node_type):
                func = getattr(self, "rename_" + node_type)
                func(current_node)

            else:
                self.rename_UNKNOWN(current_node)

        self.gen_Input()



    def _set_output_shape(self, source_node, IR_node):

        shape = graph_pb2.TensorShape()


        layer_name = source_node.name

        shape_pytorch = self.shape_dict[layer_name]


        new_dim = shape.dim.add()

        # (batch, C, H, W)  & NHWC
        if len(shape_pytorch) == 4:

            if shape_pytorch[0] == 1:
                new_dim.size = -1
            else:
                new_dim.size = shape_pytorch[0]
            for index in [2, 3, 1]:
                new_dim = shape.dim.add()
                dim = shape_pytorch[index]
                new_dim.size = dim if dim else -1
        elif len(shape_pytorch) == 2:
            if shape_pytorch[0] == 1:
                new_dim.size = -1
            else:
                new_dim.size = shape_pytorch[0]
            for _ in range(2):
                new_dim = shape.dim.add()
                new_dim.size = 1
            new_dim = shape.dim.add()
            dim = shape_pytorch[1]
            new_dim.size = dim if dim else -1


        IR_node.attr["_output_shapes"].list.shape.extend([shape])

    ##########
    # Layers #
    ##########
    def rename_UNKNOWN(self, source_node):
        print (source_node.layer)
        print (source_node.layer.data.size())
        assert False
        print("PyTorch parser has not supported operator [%s] with name [%s]."
              % (source_node.type, source_node.name))

    def gen_Input(self):
        IR_node = self.IR_graph.node.add()
        IR_node.name = 'input'
        IR_node.op = "DataInput"

        for node in self.IR_graph.node:
            if node.name in self.src_graph.input_layers:
                node.input.append('input')

        assert len(self.input_shape) == 4
        new_dim = IR_node.attr["shape"].shape.dim.add()
        if self.input_shape[0] == 1:
            new_dim.size = -1
        else:
            new_dim.size = self.input_shape[0]
        for index in [2, 3, 1]:
            new_dim = IR_node.attr["shape"].shape.dim.add()
            new_dim.size = self.input_shape[index]

        shape = graph_pb2.TensorShape()
        new_dim = shape.dim.add()
        shape_pytorch = self.input_shape

        if len(shape_pytorch) == 4:

            if shape_pytorch[0] == 1:
                new_dim.size = -1
            else:
                new_dim.size = shape_pytorch[0]
            for index in [2, 3, 1]:
                new_dim = shape.dim.add()
                dim = shape_pytorch[index]
                new_dim.size = dim if dim else -1
        elif len(shape_pytorch) == 2:
            if shape_pytorch[0] == 1:
                new_dim.size = -1
            else:
                new_dim.size = shape_pytorch[0]
            for _ in range(2):
                new_dim = shape.dim.add()
                new_dim.size = 1
            new_dim = shape.dim.add()
            dim = shape_pytorch[1]
            new_dim.size = dim if dim else -1


        IR_node.attr["_output_shapes"].list.shape.extend([shape])


    def rename_Conv(self, source_node):

        attr = source_node.attrs
        kwargs = dict()

        # dilation
        if 'dilations' in attr:
            kwargs['dilations'] = [1] + attr['dilations'] + [1]
        else:
            kwargs['dilations'] = [1] + [1, 1] + [1]

        if len(attr['pads']) == 4:
            kwargs['pads'] = [0] + attr['pads'][0:2] + [0, 0] + attr['pads'][2:] + [0]
        elif len(attr['pads']) == 2:
            kwargs['pads'] = ( [0] + attr['pads'][0:2] + [0] ) *2

        if 'strides' not in attr:
            kwargs['strides'] = [1] + [1, 1] + [1]
        else:
            kwargs['strides'] = [1] + attr['strides'] + [1]

        kwargs['group'] = attr['group']



        bias_name = '{0}.bias'.format(source_node.weights_name)
        weights_name = '{0}.weight'.format(source_node.weights_name)

        weight = self.state_dict[weights_name]

        weight = weight.numpy()
        dim = weight.ndim - 2


        IR_node = self._convert_identity_operation(source_node, new_op="Conv")
        weight = np.transpose(weight, list(range(2, dim + 2)) + [1, 0])

        self.set_weight(source_node.name, 'weights', weight)
        kwargs['kernel_shape'] = list(weight.shape)


        # handle bias
        if bias_name in self.state_dict:
            bias = self.state_dict[bias_name].numpy()
            self.set_weight(source_node.name, 'bias', bias)
            kwargs['use_bias'] = True
        else:
            kwargs['use_bias'] = False


        assign_IRnode_values(IR_node, kwargs)


    def rename_BatchNormalization(self, source_node):
        # TODO
        # output_shape

        IR_node = self._convert_identity_operation(source_node, new_op="BatchNorm")


        attr = source_node.attrs
        # epsilon
        IR_node.attr['epsilon'].f = attr['epsilon']


        bias_name = '{0}.bias'.format(source_node.weights_name)
        weights_name = '{0}.weight'.format(source_node.weights_name)
        mean_name = '{0}.running_mean'.format(source_node.weights_name)
        var_name = '{0}.running_var'.format(source_node.weights_name)



        if bias_name in self.state_dict:
            beta = self.state_dict[bias_name].numpy()
            IR_node.attr['bias'].b = True
        else:
            IR_node.attr['bias'].b = False

        if weights_name in self.state_dict:
            gamma = self.state_dict[weights_name].numpy()
            IR_node.attr['scale'].b = True
        else:
            IR_node.attr['scale'].b = False

        mean = self.state_dict[mean_name].numpy()
        variance = self.state_dict[var_name].numpy()



        if IR_node.attr['scale'].b:
            self.set_weight(source_node.name, "scale", gamma)

        if IR_node.attr['bias'].b:
            self.set_weight(source_node.name, "bias", beta)

        # mean
        self.set_weight(source_node.name, "mean", mean)

        # var
        self.set_weight(source_node.name, "var", variance)

    def rename_Relu(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op="Relu")

    def rename_Tanh(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op="Tanh")

    def rename_Sigmoid(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op="Sigmoid")

    def rename_Mul(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op="Mul")

    def rename_Maxpool(self, source_node):
        attr = source_node.attrs
        kwargs = dict()
        kwargs['strides'] = [1] + attr['strides'] + [1]
        if 'dilations' not in attr:
            kwargs['dilations'] = [1] + [1, 1] + [1]
        else:
            kwargs['dilations'] = [1] + attr['dilations'] + [1]
        kwargs['pads'] = [0] + attr['pads'][0:2] + [0, 0] + attr['pads'][2:] + [0]
        kwargs['kernel_shape'] = [1] + attr['kernel_shape'] + [1]
        IR_node = self._convert_identity_operation(source_node, new_op="Pool")

        kwargs['pooling_type'] = 'MAX'

        assign_IRnode_values(IR_node, kwargs)

    def rename_Avgpool(self, source_node):
        attr = source_node.attrs
        kwargs = dict()
        kwargs['strides'] = [1] + attr['strides'] + [1]
        if 'dilations' not in attr:
            kwargs['dilations'] = [1] + [1, 1] + [1]
        else:
            kwargs['dilations'] = [1] + attr['dilations'] + [1]
        kwargs['pads'] = [0] + attr['pads'][0:2] + [0, 0] + attr['pads'][2:] + [0]
        kwargs['kernel_shape'] = [1] + attr['kernel_shape'] + [1]
        IR_node = self._convert_identity_operation(source_node, new_op="Pool")

        kwargs['pooling_type'] = 'AVG'

        assign_IRnode_values(IR_node, kwargs)

    def rename_Flatten(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op="Flatten")

    def rename_FullyConnected(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op="FullyConnected")

        bias_name = '{0}.bias'.format(source_node.weights_name)
        weights_name = '{0}.weight'.format(source_node.weights_name)


        W = self.state_dict[weights_name].numpy().transpose()
        input_channels, output_channels = W.shape

        # Kit weight tranpose
        # weight: N x M -> C x H x W x M -> H x W x C x M -> N x M
        if self.weight_loaded:
            parent = self.src_graph.get_parent(source_node.name, [0])
            while parent.type == 'onnx::Flatten' or parent.type == 'onnx::Dropout':
                parent = self.src_graph.get_parent(parent.name, [0])
            if len(self.shape_dict[parent.name]) == 4:
                #
                original_shape = W.shape
                channel_first_list = self.shape_dict[parent.name][1:]
                dim = len(channel_first_list) + 1
                weight = W.reshape(channel_first_list + [original_shape[1]])
                assert dim > 2
                weight = weight.transpose(list(range(1, dim-1)) + [0, dim-1])
                W = weight.reshape(original_shape)

        # weights
        self.set_weight(source_node.name, 'weights', W )

        # use_bias
        if bias_name in self.state_dict:
            IR_node.attr['use_bias'].b = True
            bias = self.state_dict[bias_name].numpy()
            self.set_weight(source_node.name, 'bias', bias )
        else:
            IR_node.attr['use_bias'].b = False

        # units
        IR_node.attr['units'].i = output_channels


    def rename_Dropout(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='Dropout')
        IR_node.attr['keep_prob'].f = source_node.attrs['ratio']

    def rename_Concat(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='Concat')

        if source_node.attrs['axis'] == 1:
            IR_node.attr['axis'].i = len(self.shape_dict[source_node.name]) - 1
        else:
            IR_node.attr['axis'].i = source_node.attrs['axis']

    def rename_Add(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='Add')


    def rename_MaxPool2d(self, source_node):
        self._convert_pooling(source_node)


    def rename_View(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='Reshape')
        assign_IRnode_values(IR_node, {'shape' : list(source_node.get_attr('new_sizes'))[1:]})


    def rename_Addmm(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='FullyConnected')
        kwargs = dict()

        # handle weight
        weight = source_node.get_attr('next_functions')[2][0].next_functions[0][0].variable.data.numpy()
        weight = np.transpose(weight)
        kwargs['units'] = weight.shape[1]
        self.set_weight(source_node.name, 'weights', weight)

        # handle bias
        if source_node.get_attr('next_functions')[0][0]:
            bias = source_node.get_attr('next_functions')[0][0].variable.data.numpy()
            kwargs['use_bias'] = True
            self.set_weight(source_node.name, 'bias', weight)

        assign_IRnode_values(IR_node, kwargs)

        print(IR_node)


    ####################
    # Helper Functions #
    ####################

    @staticmethod
    def _copy_and_reop(source_node, IR_node, new_op = None):
        if new_op == None: new_op = source_node.type
        IR_node.name = source_node.name
        IR_node.op = new_op


    def _convert_identity_operation(self, source_node, in_edge_count = None, new_op = None):
        IR_node = self.IR_graph.node.add()
        PytorchParser._copy_and_reop(source_node, IR_node, new_op)
        self.convert_inedge(source_node, IR_node, 0, in_edge_count)
        self._set_output_shape(source_node, IR_node)
        return IR_node

    def _convert_pooling(self, source_node):
        kwargs = dict()
        kwargs['strides'] = [1] + list(source_node.get_attr('stride')) + [1]
        kwargs['dilations'] = [1] + list(source_node.get_attr('dilation')) + [1]
        kwargs['pads'] = ([0] + list(source_node.get_attr('padding')) + [0]) * 2
        kwargs['kernel_shape'] = [1] + list(source_node.get_attr('kernel_size')) + [1]
        IR_node = self._convert_identity_operation(source_node, new_op="Pool")

        if source_node.name.startswith('Max'):
            kwargs['pooling_type'] = 'MAX'
        elif source_node.name.startswith('Avg'):
            kwargs['pooling_type'] = 'AVG'
        else:
            raise ValueError('Unknown pooling type')

        assign_IRnode_values(IR_node, kwargs)



import sys as _sys
import argparse
import mmdnn.conversion._script.convertToIR as convertToIR
import mmdnn.conversion._script.IRToCode as IRToCode
import mmdnn.conversion._script.IRToModel as IRToModel
from six import text_type as _text_type
import uuid
import os


def _get_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '--srcFramework', '-sf',
        type=_text_type,
        choices=["caffe", "caffe2", "cntk", "mxnet", "keras", "tensorflow", 'tf', 'pytorch'],
        help="Source toolkit name of the model to be converted.")
    parser.add_argument(
        '--inputWeight', '-iw',
        type=_text_type,
        default=None,
        help='Path to the model weights file of the external tool (e.g caffe weights proto binary, keras h5 binary')
    parser.add_argument(
        '--inputNetwork', '-in',
        type=_text_type,
        default=None,
        help='Path to the model network file of the external tool (e.g caffe prototxt, keras json')
    parser.add_argument(
        '--dstFramework', '-df',
        type=_text_type,
        choices=['caffe', 'caffe2', 'cntk', 'mxnet', 'keras', 'tensorflow', 'coreml', 'pytorch', 'onnx'],
        required=True,
        help='Format of model at srcModelPath (default is to auto-detect).')
    parser.add_argument(
        '--outputModel', '-om',
        type=_text_type,
        required=True,
        help='Path to save the destination model')
    parser.add_argument(
        '--dump_tag',
        type=_text_type,
        default=None,
        help='Tensorflow model dump type',
        choices=['SERVING', 'TRAINING'])

    return parser


def _extract_ir_args(args, unknown_args, temp_filename):
    unknown_args.extend(['--srcFramework', args.srcFramework])
    if args.inputWeight is not None:
        unknown_args.extend(['--weights', args.inputWeight])
    if args.inputNetwork is not None:
        unknown_args.extend(['--network', args.inputNetwork])
    unknown_args.extend(['--dstPath', temp_filename])

    ir_parser = convertToIR._get_parser()
    return ir_parser.parse_known_args(unknown_args)


def _extract_code_args(args, unknown_args, temp_filename,network_filename):
    unknown_args.extend(['--dstFramework', args.dstFramework])
    unknown_args.extend(['--IRModelPath', temp_filename + '.pb'])
    unknown_args.extend(['--IRWeightPath', temp_filename + '.npy'])
    unknown_args.extend(['--dstModelPath', network_filename + '.py'])
    unknown_args.extend(['--dstWeightPath', temp_filename + '.npy'])
    code_parser = IRToCode._get_parser()
    return code_parser.parse_known_args(unknown_args)


def _extract_model_args(args, unknown_args, temp_filename):
    unknown_args.extend(['--framework', args.dstFramework])
    unknown_args.extend(['--inputNetwork', temp_filename + '.pb'])
    unknown_args.extend(['--inputWeight', temp_filename + '.npy'])
    unknown_args.extend(['--output', args.outputModel])
    model_parser = IRToModel._get_parser()
    return model_parser.parse_known_args(unknown_args)


def remove_temp_files(temp_filename, verbose=False):
    exts = ['.json', '.pb', '.npy', '.py']
    # exts = ['.pb', '.npy', '.py']
    for ext in exts:
        temp_file = temp_filename + ext
        if os.path.isfile(temp_file):
            os.remove(temp_file)
            if verbose:
                print('temporary file [{}] has been removed.'.format(temp_file))

def get_network_filename(framework, temp_filename, output_model_filename):
    if framework in ['pytorch']:
        return os.path.join(os.path.dirname(output_model_filename), os.path.basename(output_model_filename).split('.')[0])
    return temp_filename


def _main():
    parser = _get_parser()
    args, unknown_args = parser.parse_known_args()
    temp_filename = uuid.uuid4().hex
    ir_args, unknown_args = _extract_ir_args(args, unknown_args, temp_filename)
    ret = convertToIR._convert(ir_args)
    if int(ret) != 0:
        _sys.exit(int(ret))
    if args.dstFramework != 'coreml':
        network_filename = get_network_filename(args.dstFramework, temp_filename, args.outputModel)
        code_args, unknown_args = _extract_code_args(args, unknown_args, temp_filename, network_filename)
        ret = IRToCode._convert(code_args)
        if int(ret) != 0:
            _sys.exit(int(ret))
        from mmdnn.conversion._script.dump_code import dump_code
        dump_code(args.dstFramework, network_filename + '.py', temp_filename + '.npy', args.outputModel, args.dump_tag)
        remove_temp_files(temp_filename)

    else:
        model_args, unknown_args = _extract_model_args(args, unknown_args, temp_filename)
        ret = IRToModel._convert(model_args)
        remove_temp_files(temp_filename)
        _sys.exit(int(ret))


if __name__ == '__main__':
    _main()



import paddle.v2 as paddle

__all__ = ['alexnet']


def alexnet(input, class_dim):
    conv1 = paddle.layer.img_conv(
        input=input,
        filter_size=11,
        num_channels=3,
        num_filters=96,
        stride=4,
        padding=1)
    cmrnorm1 = paddle.layer.img_cmrnorm(
        input=conv1, size=5, scale=0.0001, power=0.75)
    pool1 = paddle.layer.img_pool(input=cmrnorm1, pool_size=3, stride=2)

    conv2 = paddle.layer.img_conv(
        input=pool1,
        filter_size=5,
        num_filters=256,
        stride=1,
        padding=2,
        groups=1)
    cmrnorm2 = paddle.layer.img_cmrnorm(
        input=conv2, size=5, scale=0.0001, power=0.75)
    pool2 = paddle.layer.img_pool(input=cmrnorm2, pool_size=3, stride=2)

    pool3 = paddle.networks.img_conv_group(
        input=pool2,
        pool_size=3,
        pool_stride=2,
        conv_num_filter=[384, 384, 256],
        conv_filter_size=3,
        pool_type=paddle.pooling.Max())

    fc1 = paddle.layer.fc(input=pool3,
                          size=4096,
                          act=paddle.activation.Relu(),
                          layer_attr=paddle.attr.Extra(drop_rate=0.5))
    fc2 = paddle.layer.fc(input=fc1,
                          size=4096,
                          act=paddle.activation.Relu(),
                          layer_attr=paddle.attr.Extra(drop_rate=0.5))

    out = paddle.layer.fc(input=fc2,
                          size=class_dim,
                          act=paddle.activation.Softmax())
    return out



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from __future__ import absolute_import
from mmdnn.conversion.examples.imagenet_test import TestKit
from mmdnn.conversion.examples.extractor import base_extractor
from mmdnn.conversion.common.utils import download_file
import paddle.v2 as paddle
import gzip
from paddle.trainer_config_helpers.config_parser_utils import \
    reset_parser


class paddle_extractor(base_extractor):

    _base_model_url = 'http://cloud.dlnel.org/filepub/?uuid='

    _image_size     = 224


    architecture_map = {
            'resnet50'             : {'params' : _base_model_url + 'f63f237a-698e-4a22-9782-baf5bb183019',},
            'resnet101'            : {'params' : _base_model_url + '3d5fb996-83d0-4745-8adc-13ee960fc55c',},
            'vgg16'                : {'params': _base_model_url + 'aa0e397e-474a-4cc1-bd8f-65a214039c2e',},

    }

    class_dim_map = {
            'resnet50'             : 1000,
            'resnet101'            : 1000,
            'vgg16'                : 1001, # work at 1001, but fail at 1000
            'alexnet'              : 1001,
    }




    @classmethod
    def dump_v2_config(cls, topology, save_path, binary=False):
        import collections

        from paddle.trainer_config_helpers.layers import LayerOutput
        from paddle.v2.layer import parse_network
        from paddle.proto import TrainerConfig_pb2
        """ Dump the network topology to a specified file.
        This function is only used to dump network defined by using PaddlePaddle V2
        API.
        :param topology: The output layers in the entire network.
        :type topology: LayerOutput|List|Tuple
        :param save_path: The path to save the dump network topology.
        :type save_path: str
        :param binary: Whether to dump the serialized network topology. The default
                    value is false.
        :type binary: bool.
        """

        if isinstance(topology, LayerOutput):
            topology = [topology]
        elif isinstance(topology, collections.Sequence):
            for out_layer in topology:
                assert isinstance(out_layer, LayerOutput), (
                    "The type of each element in the parameter topology "
                    "should be LayerOutput.")
        else:
            raise RuntimeError("Error input type for parameter topology.")

        model_str = parse_network(topology)
        with open(save_path, "w") as fout:
            if binary:
                fout.write(model_str.SerializeToString())
            else:
                fout.write(str(model_str))


    @classmethod
    def download(cls, architecture, path="./"):
        if cls.sanity_check(architecture):
            reset_parser()


            DATA_DIM = 3 * paddle_extractor._image_size * paddle_extractor._image_size  # Use 3 * 331 * 331 or 3 * 299 * 299 for Inception-ResNet-v2.
            CLASS_DIM = paddle_extractor.class_dim_map[architecture]

            image = paddle.layer.data(
                name="image", type=paddle.data_type.dense_vector(DATA_DIM))
            if 'resnet' in architecture:
                from mmdnn.conversion.examples.paddle.models import resnet
                depth = int(architecture.strip('resnet'))
                out = resnet.resnet_imagenet(image, class_dim=CLASS_DIM, depth=depth)
            elif architecture == 'vgg16':
                from mmdnn.conversion.examples.paddle.models import vgg
                out = vgg.vgg16(image, class_dim=CLASS_DIM)
            else:
                print("Not support for {} yet.", architecture)
                return None
            architecture_file = path + architecture + '.bin'
            paddle_extractor.dump_v2_config(out, architecture_file, True)

            weight_file = download_file(cls.architecture_map[architecture]['params'], directory=path, local_fname= architecture +'.tar.gz')
            if not weight_file:
                return None

            print("MXNet Model {} saved as [{}] and [{}].".format(architecture, architecture_file, weight_file))
            return (architecture_file, weight_file)


        else:
            return None


    @classmethod
    def inference(cls, architecture, files, path, image_path):

        import numpy as np
        if cls.sanity_check(architecture):
            # refer to https://github.com/PaddlePaddle/Paddle/blob/develop/python/paddle/v2/tests/test_rnn_layer.py#L35
            reset_parser()

            # refer to https://github.com/PaddlePaddle/Paddle/issues/7403
            paddle.init(use_gpu=False, trainer_count=1)

            DATA_DIM = 3 * paddle_extractor._image_size * paddle_extractor._image_size  # Use 3 * 331 * 331 or 3 * 299 * 299 for Inception-ResNet-v2.
            CLASS_DIM = paddle_extractor.class_dim_map[architecture]
            image = paddle.layer.data(
                name="image", type=paddle.data_type.dense_vector(DATA_DIM))

            if 'resnet' in architecture:
                from mmdnn.conversion.examples.paddle.models import resnet
                depth = int(architecture.strip('resnet'))
                out = resnet.resnet_imagenet(image, class_dim=CLASS_DIM, depth=depth)
            elif architecture == 'vgg16':
                from mmdnn.conversion.examples.paddle.models import vgg
                out = vgg.vgg16(image, class_dim=CLASS_DIM)
            else:
                print("Not support for {} yet.", architecture)
                return None

            _, parameters_file = files


            with gzip.open(parameters_file, 'r') as f:
                parameters = paddle.parameters.Parameters.from_tar(f)


            func = TestKit.preprocess_func['paddle'][architecture]
            img = func(image_path)
            img = np.transpose(img, [2, 0, 1])
            test_data = [(img.flatten(),)]

            predict = paddle.infer(output_layer = out, parameters=parameters, input=test_data)
            predict = np.squeeze(predict)

            return predict


        else:
            return None



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import argparse
from six import text_type as _text_type
import tensorflow as tf
from tensorflow.contrib.slim.python.slim.nets import vgg
from tensorflow.contrib.slim.python.slim.nets import inception
from tensorflow.contrib.slim.python.slim.nets import resnet_v1
from tensorflow.contrib.slim.python.slim.nets import resnet_v2
from mmdnn.conversion.examples.imagenet_test import TestKit

slim = tf.contrib.slim

input_layer_map = {
    'vgg16'         : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 224, 224, 3]),
    'vgg19'         : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 224, 224, 3]),
    'inception_v1'  : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 224, 224, 3]),
    'inception_v2'  : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 299, 299, 3]),
    'inception_v3'  : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 299, 299, 3]),
    'resnet50'      : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 299, 299, 3]),
    'resnet_v1_101' : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 224, 224, 3]),
    'resnet101'     : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 299, 299, 3]),
    'resnet152'     : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 299, 299, 3]),
    'resnet200'     : lambda : tf.placeholder(name='input', dtype=tf.float32, shape=[None, 299, 299, 3]),
}

arg_scopes_map = {
    'vgg16'         : vgg.vgg_arg_scope,
    'vgg19'         : vgg.vgg_arg_scope,
    'inception_v1'  : inception.inception_v3_arg_scope,
    'inception_v2'  : inception.inception_v3_arg_scope,
    'inception_v3'  : inception.inception_v3_arg_scope,
    'resnet50'      : resnet_v2.resnet_arg_scope,
    'resnet_v1_101' : resnet_v2.resnet_arg_scope,
    'resnet101'     : resnet_v2.resnet_arg_scope,
    'resnet152'     : resnet_v2.resnet_arg_scope,
    'resnet200'     : resnet_v2.resnet_arg_scope,
    # 'mobilenet_v1': mobilenet_v1.mobilenet_v1_arg_scope,
}

networks_map = {
    'vgg16'         : lambda : vgg.vgg_16,
    'vgg19'         : lambda : vgg.vgg_19,
    'inception_v1'  : lambda : inception.inception_v1,
    'inception_v2'  : lambda : inception.inception_v2,
    'inception_v3'  : lambda : inception.inception_v3,
    'resnet_v1_101' : lambda : resnet_v1.resnet_v1_101,
    'resnet50'      : lambda : resnet_v2.resnet_v2_50,
    'resnet101'     : lambda : resnet_v2.resnet_v2_101,
    'resnet152'     : lambda : resnet_v2.resnet_v2_152,
    'resnet200'     : lambda : resnet_v2.resnet_v2_200,
    #'mobilenet_v1' : mobilenet_v1.mobilenet_v1,
}

def _main():
    parser = argparse.ArgumentParser()

    parser.add_argument('-n', '--network', type=_text_type, help='Model Type', required=True,
        choices = input_layer_map.keys())

    parser.add_argument('-i', '--image',
        type=_text_type, help='Test Image Path')

    parser.add_argument('-ckpt', '--checkpoint',
        type=_text_type, help='Tensorflow Checkpoint file name', required=True)

    args = parser.parse_args()

    num_classes = 1000 if args.network in ('vgg16', 'vgg19', 'resnet_v1_101') else 1001

    with slim.arg_scope(arg_scopes_map[args.network]()):
        data_input = input_layer_map[args.network]()
        logits, endpoints = networks_map[args.network]()(data_input, num_classes=num_classes, is_training=False)
        labels = tf.squeeze(logits)

    init = tf.global_variables_initializer()

    with tf.Session() as sess:
        writer = tf.summary.FileWriter('./graphs', sess.graph)
        writer.close()
        sess.run(init)
        saver = tf.train.Saver()
        saver.restore(sess, args.checkpoint)
        save_path = saver.save(sess, "./imagenet_{}.ckpt".format(args.network))
        print("Model saved in file: %s" % save_path)

        if args.image:
            import numpy as np
            func = TestKit.preprocess_func['tensorflow'][args.network]
            img = func(args.image)
            img = np.expand_dims(img, axis = 0)
            predict = sess.run(logits, feed_dict = {data_input : img})
            predict = np.squeeze(predict)
            top_indices = predict.argsort()[-5:][::-1]
            result = [(i, predict[i]) for i in top_indices]
            print (result)


if __name__=='__main__':
    _main()



import base64
from google.protobuf import json_format
from importlib import import_module
import json
import numpy as np
import os
import sys

from mmdnn.conversion.caffe.errors import ConversionError
from mmdnn.conversion.caffe.common_graph import fetch_attr_value
from mmdnn.conversion.caffe.utils import get_lower_case, get_upper_case, get_real_name


class JsonFormatter(object):
    '''Dumpt a DL graph into a Json file.'''

    def __init__(self, graph):
        self.graph_def = graph.as_graph_def()

    def dump(self, json_path):
        json_txt = json_format.MessageToJson(self.graph_def)
        parsed = json.loads(json_txt)
        formatted = json.dumps(parsed, indent=4, sort_keys=True)
        with open(json_path, 'w') as f:
            f.write(formatted)

    
class PyWriter(object):
    '''Dumpt a DL graph into a Python script.'''

    def __init__(self, graph, data, target):
        self.graph = graph
        self.data = data
        self.tab = ' ' * 4
        self.prefix = ''
        target = target.lower()
        if target == 'tensorflow':
            self.target = target
            self.net = 'TensorFlowNetwork'
        elif target == 'keras':
            self.target = target
            self.net = 'KerasNetwork'
        elif target == 'caffe':
            self.target = target
            self.net = 'CaffeNetwork'
        else:
            raise ConversionError('Target %s is not supported yet.' % target)

    def indent(self):
        self.prefix += self.tab

    def outdent(self):
        self.prefix = self.prefix[:-len(self.tab)]

    def statement(self, s):
        return self.prefix + s + '\n'

    def emit_imports(self):
        return self.statement('from dlconv.%s import %s\n' % (self.target, self.net))

    def emit_class_def(self, name):
        return self.statement('class %s(%s):' % (name, self.net))

    def emit_setup_def(self):
        return self.statement('def setup(self):')

    def emit_node(self, node):
        '''Emits the Python source for this node.'''
        
        def pair(key, value):
            return '%s=%s' % (key, value)
        args = []
        for input in node.input:
            input = input.strip().split(':')
            name = ''.join(input[:-1])
            idx = int(input[-1])
            assert name in self.graph.node_dict
            parent = self.graph.get_node(name)
            args.append(parent.output[idx])
        #FIXME:
        output = [node.output[0]]
        # output = node.output
        for k, v in node.attr:
            if k == 'cell_type':
                args.append(pair(k, "'" + fetch_attr_value(v) + "'"))
            else:
                args.append(pair(k, fetch_attr_value(v)))
        args.append(pair('name', "'" + node.name + "'")) # Set the node name
        args = ', '.join(args)
        return self.statement('%s = self.%s(%s)' % (', '.join(output), node.op, args))

    def dump(self, code_output_dir):
        if not os.path.exists(code_output_dir):
            os.makedirs(code_output_dir)
        file_name = get_lower_case(self.graph.name)
        code_output_path = os.path.join(code_output_dir, file_name + '.py')
        data_output_path = os.path.join(code_output_dir, file_name + '.npy')
        with open(code_output_path, 'w') as f:
            f.write(self.emit())
        with open(data_output_path, 'wb') as f:
            np.save(f, self.data)
        return code_output_path, data_output_path

    def emit(self):
        # Decompose DAG into chains
        chains = []
        for node in self.graph.topologically_sorted():
            attach_to_chain = None
            if len(node.input) == 1:
                parent = get_real_name(node.input[0])
                for chain in chains:
                    if chain[-1].name == parent: # Node is part of an existing chain.
                        attach_to_chain = chain
                        break
            if attach_to_chain is None: # Start a new chain for this node.
                attach_to_chain = []
                chains.append(attach_to_chain)
            attach_to_chain.append(node)
            
        # Generate Python code line by line
        source = self.emit_imports()
        source += self.emit_class_def(self.graph.name)
        self.indent()
        source += self.emit_setup_def()
        self.indent()
        blocks = []
        for chain in chains:
            b = ''
            for node in chain:
                b += self.emit_node(node)
            blocks.append(b[:-1])
        source += '\n\n'.join(blocks)
        return source


class ModelSaver(object):

    def __init__(self, code_output_path, data_output_path):
        self.code_output_path = code_output_path
        self.data_output_path = data_output_path

    def dump(self, model_output_dir):
        '''Return the file path containing graph in generated model files.'''
        if not os.path.exists(model_output_dir):
            os.makedirs(model_output_dir)
        sys.path.append(os.path.dirname(self.code_output_path))
        file_name = os.path.splitext(os.path.basename(self.code_output_path))[0]
        module = import_module(file_name)
        class_name = get_upper_case(file_name)
        net = getattr(module, class_name)
        return net.dump(self.data_output_path, model_output_dir)


class GraphDrawer(object):

    def __init__(self, toolkit, meta_path):
        self.toolkit = toolkit.lower()
        self.meta_path = meta_path

    def dump(self, graph_path):
        if self.toolkit == 'tensorflow':
            from dlconv.tensorflow.visualizer import TensorFlowVisualizer
            if self._is_web_page(graph_path):
                TensorFlowVisualizer(self.meta_path).dump_html(graph_path)
            else:
                raise NotImplementedError('Image format or %s is unsupported!' % graph_path)
        elif self.toolkit == 'keras':
            from dlconv.keras.visualizer import KerasVisualizer
            png_path, html_path = (None, None)
            if graph_path.endswith('.png'):
                png_path = graph_path
            elif self._is_web_page(graph_path):
                png_path = graph_path + ".png"
                html_path = graph_path
            else:
                raise NotImplementedError('Image format or %s is unsupported!' % graph_path)
            KerasVisualizer(self.meta_path).dump_png(png_path)
            if html_path:
                self._png_to_html(png_path, html_path)
                os.remove(png_path)
        else:
            raise NotImplementedError('Visualization of %s is unsupported!' % self.toolkit)

    def _is_web_page(self, path):
        return path.split('.')[-1] in ('html', 'htm')

    def _png_to_html(self, png_path, html_path):
        with open(png_path, "rb") as f:
            encoded = base64.b64encode(f.read()).decode('utf-8')
        source = """<!DOCTYPE>
<html>
    <head>
        <meta charset="utf-8">
        <title>Keras</title>
    </head>
    <body>
        <img alt="Model Graph" src="data:image/png;base64,{base64_str}" />
    </body>
</html>""".format(base64_str=encoded)
        with open(html_path, 'w', encoding='utf-8') as f:
            f.write(source)


#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

from mmdnn.conversion.common.DataStructure.graph import GraphNode, Graph
from tensorflow.core.framework.node_def_pb2 import NodeDef
from tensorflow.core.framework import attr_value_pb2


class TensorflowGraphNode(GraphNode):

    def __init__(self, layer):
        super(TensorflowGraphNode, self).__init__(layer)
        self.in_nodes = list()
        self.out_nodes = list()
        self._scope = str()


    @property
    def scope(self):
        return self._scope

    @scope.setter
    def scope(self, scope):
        self._scope = scope


    @property
    def name(self):
        return self.layer.name


    @property
    def type(self):
        return self.layer.op


    @property
    def tf_layer(self):
        return self.layer


    def get_attr(self, name, default_value = None):
        if name in self.layer.attr:
            attr = self.layer.attr[name]
            field = attr.WhichOneof('value')
            val = getattr(attr, field) if field else default_value
            if isinstance(val, attr_value_pb2.AttrValue.ListValue):
                return list(val.ListFields()[0][1])
            else:
                return val.decode('utf-8') if isinstance(val, bytes) else val
        else:
            return default_value


class TensorflowGraph(Graph):

    multi_tensor_type = [
        "Slice",
        "Split",
        "Unpack"
    ]


    def __init__(self, model):
        # sanity check.
        pass

        super(TensorflowGraph, self).__init__(model)
        self.model = model


    def build(self):
        for i, layer in enumerate(self.model.node):
            self.layer_map[layer.name] = TensorflowGraphNode(layer)
            self.layer_name_map[layer.name] = layer.name
            for pred in layer.input:
                if pred not in self.layer_map:
                    if not pred.split(':')[0] in self.layer_map: #test
                        new_node = NodeDef()
                        new_node.name = pred
                        new_node.op = "NoOp"
                        self.layer_map[pred] = TensorflowGraphNode(new_node)
                        self.layer_name_map[pred] = pred

                self.tf_make_connection(pred, layer.name)

        super(TensorflowGraph, self).build()


    def tf_make_connection(self, src, dst):

        if ':' not in src and self.get_node(src).type in self.multi_tensor_type:
            src += ':0'

        self._make_connection(src, dst)
        src_node = self.get_node(src.split(':')[0])
        dst_node = self.get_node(dst.split(':')[0])

        if not src_node in self.layer_map[dst.split(':')[0]].in_nodes:
            self.layer_map[dst.split(':')[0]].in_nodes.append(src_node)
        if not dst_node in self.layer_map[src.split(':')[0]].out_nodes:
            self.layer_map[src.split(':')[0]].out_nodes.append(dst_node)
